def running_mean(y, N):
    #if N is even window is not centered on points but at half between. One point less is returned.  
    y2=np.insert(y, 0, 0)
    cumsum = y2*0+np.nan_to_num(y2).cumsum()
    cumsum= (cumsum[N:] - cumsum[:-N]) / N
    start=np.array([np.nanmean(y[:i*2+1]) for i in range(N)][:int(N/2)])  #first elementh of cumsum is the average 
    end=np.array([np.nanmean(y[-(i*2+1):]) for i in range(N)][:int(N/2)][::-1])
    cumsum= np.hstack([start,cumsum,end]) #add ends
    return cumsum


import numpy as np
from dataIO.span import span as span # rename to span to avoid conflict with argument.
from typing import Callable
import warnings
import pdb

class EmptyRangeWarning(RuntimeWarning):
    pass
    #def __init__(self, *args, **kwargs):
    #    super().__init__( *args, **kwargs)

def remove_outliers(data: np.array,
                    nsigma: float = 3.,
                    itmax: int = 1,
                    flattening_func: Callable[[np.array],np.array] = None,
                    span: bool = False, #deprecated
                    print_partial: bool = False) -> np.array:
                    
    """ iteratively remove outliers out of an interval. Returns a mask, True on data to keep.
        
        For each iteration, outliers outside `nsigma` standard deviations from average are removed. Loop ends after `itmax` (default 1) iterations or if convergency is reached (two consecutive iterations with same stdev).
        A flattening function (or any function that returns a modified version of data) can be passed to be performed at each iteration. At each interaction function is calculated on accepted values only.
        If at any point data is empty, an empty array is returned.
        
        A mask is returned by default, A copy of the original array with invalid values as nan. Values array can equivalently be obtained from mask by `np.where(mask,a,np.nan)`.
        
        `span` argument is deprecated and it will be removed. Please update your code to use `from dataIO.span import span; span (remove_outliers(data,...))`.

        """
        #see also dataIO.span.filtered_span
        
    if span: # useless and was giving conflict, solved by workaround below.
        print("`span` argument is deprecated and it will be removed. Please update your code to use `from dataIO.span import span; span (remove_outliers(data,...))`.\n\nentering debugger, `c` to continue, `u` to see caller function, `l` list code.")
        pdb.set_trace()
    get_span = span  #rename variable
    from dataIO.span import span
    
    if flattening_func is not None: data=flattening_func(data)
    
    M = np.isfinite(data)  #mask for good data
    sigma=np.nanstd(data[M])
    sigmast=0
    i=1
    if itmax > 0:
        while sigma != sigmast:
            sigmast=sigma
            # keep in M only points in the range and recalculate sigma
            M = M & (np.abs(data-np.nanmean(data)) < (nsigma*sigma))
            data = np.where(M,data,np.nan)
            if flattening_func is not None: data=flattening_func(data)
            sigma=np.nanstd(data)
            if print_partial:
                print (i,sigma,sigmast)        
            if i >= itmax or np.all(M is False):
                break                
            i=i+1
    elif print_partial:
        print ("itmax = 0, just M valid data.")
    
    if not (M).any():
        warnings.warn('Returning empty array after filtering of outliers.',EmptyRangeWarning)
        return []
    #pdb.set_trace()  
    #if get_span: return span(data)
    #return M if mask else np.where(M,data,np.nan)
    return M


def filter_outliers(data: np.array,
                    nsigma: float = 3.,
                    itmax: int = 1,
                    flattening_func: Callable[[np.array],np.array] = None,
                    span: bool = False, #deprecated
                    print_partial: bool = False) -> np.array:
                    
    """ This is a generator version that uses remove_outliers to provide iterative result if 
    that iteratively remove outliers out of an interval. 
    Each iteration return a mask, True on data to keep.
        
        For each iteration, outliers outside `nsigma` standard deviations from average are removed. Loop ends after `itmax` (default 1) iterations or if convergency is reached (two consecutive iterations with same stdev).
        A flattening function (or any function that returns a modified version of data) can be passed to be performed at each iteration. At each interaction function is calculated on accepted values only.
        If at any point data is empty, an empty array is returned.
        
        A mask is returned by default, A copy of the original array with invalid values as nan. Values array can equivalently be obtained from mask by `np.where(mask,a,np.nan)`.
        
        `span` argument is deprecated and it will be removed. Please update your code to use `from dataIO.span import span; span (remove_outliers(data,...))`.

        """
        #see also dataIO.span.filtered_span
        
    if span: # useless and was giving conflict, solved by workaround below.
        print("`span` argument is deprecated and it will be removed. Please update your code to use `from dataIO.span import span; span (remove_outliers(data,...))`.\n\nentering debugger, `c` to continue, `u` to see caller function, `l` list code.")
        pdb.set_trace()
    get_span = span  #rename variable
    from dataIO.span import span
    
    # --------------
    if flattening_func is not None: data=flattening_func(data)
    
    M = np.isfinite(data)  #mask for good data
    sigma=np.nanstd(data[M])
    sigmast=0
    i=1
    if itmax >= 0:
        while sigma != sigmast:
            if i >= itmax or np.all(M is False):
                break 
            sigmast=sigma
            # keep in M only points in the range and recalculate sigma
            M = M & (np.abs(data-np.nanmean(data)) < (nsigma*sigma))
            data = np.where(M,data,np.nan)
            if flattening_func is not None: data=flattening_func(data)
            sigma=np.nanstd(data)
            if not (M).any():
                warnings.warn('Returning empty array after filtering of outliers.',EmptyRangeWarning)
                yield []  
            yield M      
               
            i=i+1

def test_filter_outliers(a=None):
    if a is None:
        a=np.arange(10)
    
    print('initial array',a,'\n--------------')    
    for i,m in enumerate(filter_outliers(a,nsigma=1,itmax=3)):
        print('iter: ',i)
        print('mask: ',m)
        print('values: ',a[m])
        print('\n')
    
def test_remove_outliers(a=None):
    if a is None:
        a=np.arange(10)
    
    print('initial array',a,'\n--------------')
    
    b=remove_outliers(a,nsigma=1)
    print('span=True,nsigma=1',b,'\n--------------')
    
    b=remove_outliers(a,nsigma=1,itmax =3,print_partial=True)
    print('span=True,nsigma=1,itmax =3,print_partial=True',b,'\n--------------')

    b=remove_outliers(a,nsigma=3)
    print('nsigma=3',b,'\n--------------')
    
    b=remove_outliers(a,nsigma=0)
    print('nsigma=0',b,'\n--------------')    

if __name__ == "__main__":
    a=np.arange(10)
    test_remove_outliers(a)


# -*- coding: utf-8 -*-
"""
Created on Fri Jun  8 00:36:57 2018

@author: Vincenzo
"""

import matplotlib.pyplot as plt
import numpy as np
from PIL import ImageFont
from PIL import ImageDraw
from PIL import Image
import pdb 

def original_make_caption():
    if mode=='tagcode': #single string (e.g. for filenames)
        #teh string manipulation is to avoid leading and ending underscores if head or tail are not set
        novo = ((' '.join([head,exp,foc,"".join(re.findall('\d+',fn)),tail])).strip()).replace(" ","_")
        #novo = '_'.join([head,exp,foc,"".join(re.findall('\d+',fn)),tail]) #the re gives all digits in filename, I believe.
    elif mode=='plottitle':
        novo="%s Exp:%s,f/n:%s %s"%(fn,exp,foc,dt)
    elif mode=='multiline': #caption as multiline text
        #texto is a single string with spaces between words to be splitted in lines
        texto=" ".join([head,fn,"exp:%s f#:%s (EV%1.1f) %i X %i %s"%(exp,foc,calculate_EV(self.foc,self.exp,self.iso),abs(int(yl[1]-yl[0])),abs(int(xl[1]-xl[0])),dt),"offset:%s,%s (or. size:%s,%s)"%(int(xl[0]),int(yl[0]),xsize,ysize),tail])
        novo = textwrap.wrap(texto, width=textwidth)
    else:
        raise ValueError('Mode not recognized %s'%mode)

    return novo   
        
def change_labels(newlab):
    """Change labels of an existing legend modifying line labels."""
    
    for ll,c in zip(plt.gca().lines,newlab):
        ll.set_label(c)
    return plt.legend()
        
def legendbox(text,ax=None,loc=0,clear = False, *args,**kwargs):
    """Create a legend from class matplotlib.legend.Legend and add it to axis.
        `text` is a list of strings to be passed to `labels` parameter in Legend.
        Unclear how line styles are set (probably none).
    if clear is set to True, remove previous legends.
        """
    '''
    N.B.: this is not useful any more. Not sure if it was added or just I didn't know about it, but the same can be obtained with plt.legend() (adding the old legend to axis, artists), or alternatively this function can be readapted in that sense.

        The original function was written to add legends without lines (e.g. stats boxes in surface plots)
        without erasing previous legends (only one legend could be built per axis).
        
        It might also be useful to use this function to modify label text but not lines labels (not sure if it is of any use).
        https://matplotlib.org/stable/tutorials/intermediate/legend_guide.html
    '''
    
    from matplotlib.patches import Patch
    from matplotlib.legend import Legend
    #------ SAME BUT WITH CUSTOM LEGEND (DOESN'T ERASE OTHER LEGENDS)

    #plt.plot([1,2],[2,3],label='cane')
    #plt.legend(loc=1) 
    
    #pdb.set_trace()
    if np.size(text) == 1 and not isinstance(text, list): text = [text]
    if ax is None:
        ax=plt.gca()
    if clear:
        ax.get_legend().remove()
    p = [Patch(label=ll,visible=False) for ll in text]
    leg = Legend(ax,p,text,loc=loc,handletextpad=0, handlelength=0,
                 *args,**kwargs)
    ax.add_artist(leg)

    return leg
    
def test_legendbox():
    from pySurf.testSurfaces import make_range
    from pySurf.data2D import plot_data, get_stats
    
    plt.clf()
    nx=4
    ny=3
    data,x,y=make_range(nx,ny)
    plot_data(data,x,y)
    l = legendbox(get_stats(data,units=['mm','mm','um']))
    plt.show()
    return l
    
"""From imageCaption (A module that creates a caption as image):"""

def autofont(novo,csize,minfont=10):
    """Given a text novo (as list of lines), determine the font size that fits an image of size csize."""
    linefraction=0.1 #vertical margin for lines as a fraction of line height
    
    fontsize=minfont
    font = ImageFont.truetype("arial.ttf", fontsize)
    longestline=novo[np.argmax([len(n) for n in novo])]
    while font.getsize(longestline)[0] < csize[0]:
        # iterate until the text size is just smaller than the criteria
        fontsize += 1
        font = ImageFont.truetype("arial.ttf", fontsize)
    # optionally de-increment to be sure it is less than criteria
    fontsize -= 1 
    font = ImageFont.truetype("arial.ttf", fontsize)       
    
    #decrease font size if it doesn't fit 
    while font.getsize(longestline)[1]*len(novo)*(1+linefraction) > csize[1]:
        # iterate until the text size is just larger than the criteria
        fontsize -= 1
        font = ImageFont.truetype("arial.ttf", fontsize)
    # optionally de-increment to be sure it is less than criteria
    fontsize -= 1 
    font = ImageFont.truetype("arial.ttf", fontsize)
    return font
    
def render_caption(novo,csize,minfont=10):
    """return a rendered image of the text novo (as list of lines) determining the font size that fits the box. csize is the size of the returned image (x,y) in pixels."""

    offset = 10  #margin in pixels between linesheight   
    font=autofont(novo,csize,minfont)
    capim=Image.new("RGB",csize)
    d = ImageDraw.Draw(capim)
    #d = ImageDraw.Draw(newim)
    for line in novo:
        d.text((0,offset),line,(255,255,255),font=font)
        offset += font.getsize(line)[1]
    del d
    
    return capim    
    
if __name__ == "__main__":
    l = test_legendbox()



"""
Extensible vectorizable list.

In final version has list properties `iterators`,
`aggregator`,`expanded` containing methods that are vectorized in different ways. Iterator is default, return a list with each element obtained by applying its method with same name to each element of `superlist`. 
Aggregator accumulate a binary operation, returning a single value. Expanded return a single element by passing the content of `superlist` as n elements to the element method. 
Note that passing parameters must be handled accordingly. In particular a list of parameters or a single value can be passed to iterators, however this can create ambiguity. For example, an element method can accept list or single values, so it is not clear when a list is passed to superlist, if this means that the same list is passed to all elements as list argument of element method, rather than passing each element of list argument as scalar argument of elements method. There is no ambiguity if a list of lists or a list with len different from number of superlist elements, this should be checked, otherwise a warning should be visualized telling about the possible ambiguity and default action (possibly passing one value per element, call with nested list [[element]] to apply as list to all values).

In this experimental version, different ways to access object methods are tested.

functions operating on a list of Data2D objects

Another problem is to make sure that superlists are returned on list operators like slices, how to do this? Metaclasses?

2022/07/12 after last changes it seems to work quite well. 
TODO: implement assignment, e.g. pl.x = 12.398/pl.x

2020/05/26 moved to dataIO"""

# turned into a class derived from list 2020/01/16, no changes to interface,
# everything should work with no changes.

import os, sys
import matplotlib.pyplot as plt
import numpy as np
import pdb

class xSuperlist(list):
    """A list of pySurf.Data2D objects on which unknown operations are performed serially."""    
    # property returns a function, which if called returns a list with the property of each item
    # method returns a list with the methods of each item as function. 
    
    def __getattr__(self,name,*args,**kwargs): 
            # devo costruire una nuova funzione che preso un oggetto
            # lista ritorna un oggetto lista ottenuto dal valore restituito dalla funzione su ogni elemento.
            def newfunc(*args, **kwargs):
                attr = [object.__getattribute__(name) for object in self]
                result = [a(*args, **kwargs) if hasattr(attr, '__call__') else a for a in attr]
                return result
            return newfunc
            
class Superlist(list):
    """Working version from Superlist6. """   
    # A problem is that a list is returned, not a superlist, so for example calling
    #   s.crop([2,10]) returns a list of objects, but I cannot apply a second time about
    #   method. 
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
    def __getattr__(self,name,*args,**kwargs): 
        attr = [object.__getattribute__(name) for object in self]
        
        if hasattr(attr[0], '__call__'):
            # builds a function that, taken a list object oggetto
            # return a list of the results obtained applying function to each item.
            def newfunc(*args, **kwargs):
                attr = [object.__getattribute__(name) for object in self]
                result = [a(*args, **kwargs) for a in attr]
                return result
            return newfunc
        else:
            # return list of attributes
            return attr
    
    def __getattribute__(self,name):
        attr = object.__getattribute__(self, name)
        if hasattr(attr, '__call__'):
            def newfunc(*args, **kwargs):
                result = attr(*args, **kwargs)
                return result
            return newfunc
        else:
            return attr
            

## DEVELOPMENT VERSIONS

class superlist1(list):
    """Test class that vectorizes methods. Basic, implements properties,
        but not methods."""
    
    def __getattr__(self,name):  #originariamente usava __getattribute__, che riceve attributo
        # prima di chiamarlo (quindi anche se gia' esistente).
        #attr = [obj.__getattr__(self, name) for obj in self] #questo non funziona
        attr = [object.__getattribute__(name) for object in self]
        
        result = []
        if hasattr(attr, '__call__'):   #try to identify a method as a callable attr
            def newfunc(*args, **kwargs):
                #pdb.set_trace()
                print('before calling %s' %a.__name__)
                result = a(*args, **kwargs)
                print('done calling %s' %a.__name__)
                return result
            return newfunc
        else:
            return attr

        return result  #it works as a property. As method, this returns a list of methods, but since the method is called as sl.method(),
        #gives an error as it tries to call the list.
        # deve ritornare una funzione che ritorna una lista.

class superlist2(list):
    """Test class that vectorizes methods."""
    
    def __getattr__(self,name):  #originariamente usava __getattribute__, che riceve attributo
        # prima di chiamarlo (quindi anche se gia' esistente).
        #attr = [obj.__getattr__(self, name) for obj in self] #questo non funziona
        attr = [object.__getattribute__(name) for object in self]
        
        """
        if hasattr(attr[0], '__call__'):
            attr=attr[0]
            def newfunc(*args, **kwargs):
                #pdb.set_trace()
                print('before calling %s' %attr.__name__)
                result = attr(*args, **kwargs)
                print('done calling %s' %attr.__name__)
                return result
            return newfunc
        else:
            return [a for a in attr]
        """
        
        result = []
        for a in attr:
            #print('loop ',a)
            #pdb.set_trace()
            if hasattr(a, '__call__'):
                result=np.vectorize(a)
                """
                def newfunc(*args, **kwargs):
                    #pdb.set_trace()
                    print('before calling %s' %a.__name__)
                    result = a(*args, **kwargs)
                    print('done calling %s' %a.__name__)
                    return result
                result.append(newfunc)
                """
            else:
                result.append(a)

        return result  #it works as a property. As method, this returns a list of methods, but since the method is called as sl.method(),
        #gives an error as it tries to call the list.
        # deve ritornare una funzione che ritorna una lista.
        
class superlist3(list):
    """Test class that vectorizes methods.  """
    
    def __getattr__(self,name):  #originariamente usava __getattribute__, che riceve attributo
        # prima di chiamarlo (quindi anche se gia' esistente).
        #attr = [obj.__getattr__(self, name) for obj in self] #questo non funziona
        attr = [object.__getattribute__(name) for object in self]
        
        result = []
        for a in attr:
            if hasattr(a, '__call__'):
                result=np.vectorize(a)
            else:
                result.append(a)

        return result  #it works as a property. As method, this returns a list of methods, but since the method is called as sl.method(),
        #gives an error as it tries to call the list.
        # deve ritornare una funzione che ritorna una lista.  

class superlist4(list):
    """Test class that vectorizes methods.  """
    
    def __getattr__(self,name):  #originariamente usava __getattribute__, che riceve attributo
        # prima di chiamarlo (quindi anche se gia' esistente).
        #attr = [obj.__getattr__(self, name) for obj in self] #questo non funziona
        attr = [object.__getattribute__(name) for object in self]
        
        result = []
        for a in attr:
            if hasattr(a, '__call__'):
                result=np.vectorize(a)
            else:
                result.append(a)

        return superlist4(result)  #deve restituire un istanza della classe
                #stessa, perche' se ritorna lista non posso applicare ulteriori
                #metodi.

class superlist_alt(list):
    """2021/07/14 da https://stackoverflow.com/questions/2704434/intercept-method-calls-in-python anche testata in Dlist."""
    
    def __getattribute__(self,name):
        attr = object.__getattribute__(self, name)
        if hasattr(attr, '__call__'):
            def newfunc(*args, **kwargs):
                print('before calling %s' %attr.__name__)
                result = attr(*args, **kwargs)
                print('done calling %s' %attr.__name__)
                return result
            return newfunc
        else:
            return attr
            
class superlist5(list):
    """2021/07/14 da Dlist, versione funzionante con proprieta' e metodi, portata qui e non testata."""
    
    def __getattr__(self,name,*args,**kwargs):  #originariamente usava __getattribute__, che riceve attributo
        # prima di chiamarlo (quindi anche se gia' esistente).
        #attr = object.__getattr__(self, name) #questo non funziona
        
        attr = [object.__getattribute__(name) for object in self]
        print("ATTR:",attr,*args,**kwargs)
        print("SELF,name",self,name)
        
        if hasattr(attr[0], '__call__'):
            print("call methods of items")
            # devo costruire una nuova funzione che preso un oggetto
            # lista ritorna un oggetto lista ottenuto dal valore restituito dalla funzione su ogni elemento.
            def newfunc(*args, **kwargs):
                attr = [object.__getattribute__(name) for object in self]
                result = [a(*args, **kwargs) for a in attr]
                return result
            return newfunc
            
            """
            # per quale motivo restituire una funzione?

                if hasattr(attr[0], '__call__'):
                    return [a(*args,**kwargs) for a in attr]  # e' una generica lista 
                else:
                    # return list of attributes
                    return Dlist(attr) #attr

            Il codice quasi funziona, nel senso che genera plot, ma da anche errore. In realta' chiamando D.plot  (senza parentesi), la lista viene restituita comunque.
            Questo perche' __getattr__ viene chiamato da qualcosa che si aspetta di vedere restituita una funzione, non un valore.

            Per questo l'equivalente unidimensionale era:
            
            def newfunc(*args, **kwargs):
                #pdb.set_trace()
                #print('before calling %s' %attr.__name__)
                result = attr(*args, **kwargs)
                #print('done calling %s' %attr.__name__)
                return result
            return newfunc
            """
        else:
            # return list of attributes
            print("return atribute of items")
            return superlist5(attr) #attr  #this can also be list. using superlist anyway makes it possible to call item methods also on results.
            # example, I can call shape of shapes from nested arrays. [[1,2],[1,2,3],[4,5]].shape.shape -> [2,2,2] (not working)
            
class Superlist6(list):
    """Working version for debug of getattr/getattribute.
    print messages."""    
    
    def __getattr__(self,name,*args,**kwargs): 
        print("GETATTR!")
        attr = [object.__getattribute__(name) for object in self]
        print("ATTR:",attr,*args,**kwargs)
        print("SELF,name",self,name)
        
        if hasattr(attr[0], '__call__'):
            print("call methods of items")
            # devo costruire una nuova funzione che preso un oggetto
            # lista ritorna un oggetto lista ottenuto dal valore restituito dalla funzione su ogni elemento.
            def newfunc(*args, **kwargs):
                attr = [object.__getattribute__(name) for object in self]
                result = [a(*args, **kwargs) for a in attr]
                return result
            return newfunc
        else:
            # return list of attributes
            print("return atribute of items")
            return attr
    
    def __getattribute__(self,name):
        print("GETATTRIBUTE!")
        attr = object.__getattribute__(self, name)
        if hasattr(attr, '__call__'):
            def newfunc(*args, **kwargs):
                print('before calling %s' %attr.__name__)
                result = attr(*args, **kwargs)
                print('done calling %s' %attr.__name__)
                return result
            return newfunc
        else:
            return attr

class Superlist(list):
    """Working version from Superlist6. Returns a superlist."""   
    # A problem was that if a list is returned, not a superlist, for example calling
    #   s.crop([2,10]) returns a list of objects, but I cannot apply a second time about
    #   method. 
    
    
    def __getattr__(self,name,*args,**kwargs): 
        attr = [object.__getattribute__(name) for object in self]
        
        if hasattr(attr[0], '__call__'):
            # builds a function that, taken a list object oggetto
            # return a list of the results obtained applying function to each item.
            def newfunc(*args, **kwargs):
                attr = [object.__getattribute__(name) for object in self]
                result = Superlist()
                result.extend([a(*args, **kwargs) for a in attr])
                return result
            return newfunc
        else:
            # return list of attributes
            return attr
    
    def __getattribute__(self,name):
        attr = object.__getattribute__(self, name)
        if hasattr(attr, '__call__'):
            def newfunc(*args, **kwargs):
                result = attr(*args, **kwargs)
                return result
            return newfunc
        else:
            return attr


def test_superlist_dlist(D):
    #create a dlist D
    # test
    print('dlist:')
    print(D)
    print('\ntest property (D.name):')
    print(D.name)
    print('\ntest method (D.plot):')
    print(D.plot())
    
    # existing method
    D.plot

    # existing method call
    D.plot()
      
    # non existing 
    D.cane


         
def test_superlist(cls,obj=None):

    print ('test class ',cls)
    if obj is None:
        #s=type(cls,[np.arange(4),np.arange(3)])
        s = cls([np.arange(4),np.arange(3)])
        #s = superlist([np.arange(4),np.arange(3)])

    print('original data:')
    print(s)
    print('\ntest property (np.shape):')
    print(s.shape)
    print('\ntest method (np.flatten):')
    print(s.flatten())
    return s
    
    
if __name__ == "__main__":
    s = test_superlist(Superlist)
        #test_superlist(superlist)
    """
    except: # catch *all* exceptions
        e = sys.exc_info()[0]
        print( "Error: %s" % e )
    """


import inspect  # to use with importing docstring

def update_docstring(func,source,delimiter=''):
    """given a current function and a source function, update current function docstring
     appending signature and docstring of source.

     It provides a decorator @update_docstring(source) update the decorated
       function docstring. User should take care to create docstring of decorated function
       only as preamble to source docstrings, e.g.:
       this is func function, derived from source function by modifying the usage of parameter foo.
       parameter foo is changed of sign."""
       
       
    doc0="" if func.__doc__ is None else func.__doc__
    func.__doc__='\n'.join([doc0,delimiter,source.__name__+str(inspect.signature(source)),source.__doc__])
    return func



"""
TODO: check if imports winsound (or similar libraries for other OSs,
returns a beep function with the best available beep (ignore extra parameters)

import winsound
frequency = 2500  # Set Frequency To 2500 Hertz
duration = 1000  # Set Duration To 1000 ms == 1 second
winsound.Beep(frequency, duration)
"""

def beep(*args,**kwargs):
    """Emits a sound (useful if you want to recall attention after a long run and you cannot remember the character code for it)."""
    print('\a')
    
    



# From MandlebrotDisplay, I remove the class to adjust it to my example.
#http://matplotlib.org/examples/event_handling/viewlims.html

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
from dataIO.span import span
import copy

def mandel(xstart, xend, ystart, yend):
    niter=50
    radius=2.
    power=2
    dims = plt.gca().axesPatch.get_window_extent().bounds
    width = int(dims[2] + 0.5)
    height = int(dims[2] + 0.5)
    x = np.linspace(xstart, xend, width)
    y = np.linspace(ystart, yend, height).reshape(-1, 1)
    c = x + 1.0j * y
    threshold_time = np.zeros((height, width))
    z = np.zeros(threshold_time.shape, dtype=np.complex)
    mask = np.ones(threshold_time.shape, dtype=np.bool)
    for i in range(niter):
        z[mask] = z[mask]**power + c[mask]
        mask = (np.abs(z) < radius)
        threshold_time += mask
    return threshold_time
        
        
def ax_update(ax):
    ax.set_autoscale_on(False)  # Otherwise, infinite loop

    # Get the number of points from the number of pixels in the window
    #dims = ax.axesPatch.get_window_extent().bounds

    # Get the range for the new area
    xstart, ystart, xdelta, ydelta = ax.viewLim.bounds
    xend = xstart + xdelta
    yend = ystart + ydelta

    # Update the image object with our new data and extent
    xl=[xstart,xend]
    yl=[ystart,yend]

    # Update the image object with our new data and extent
    im = ax.images[-1]
    cdata=mandel(xstart, xend, ystart, yend)
    im.set_data(cdata)
    crange=span(cdata)
    im.set_extent((xstart, xend, ystart, yend))
    if not (im.colorbar is None):
        fig=ax.get_figure()
        s = copy.copy( fig.canvas.toolbar._views )  
        #this store and restore is made to preserve the zooming history, 
        p = copy.copy( fig.canvas.toolbar._positions )
        im.colorbar.remove()
        plt.clim(crange)
        plt.colorbar()
        fig.canvas.toolbar._views = s
        fig.canvas.toolbar._positions = p
    #
        
    #ax.figure.canvas.draw_idle()

def ax_update2(ax):
    ax.set_autoscale_on(False)  # Otherwise, infinite loop

    # Get the number of points from the number of pixels in the window
    #dims = ax.axesPatch.get_window_extent().bounds

    # Get the range for the new area
    xstart, ystart, xdelta, ydelta = ax.viewLim.bounds
    xend = xstart + xdelta
    yend = ystart + ydelta

    # Update the image object with our new data and extent
    xl=[xstart,xend]
    yl=[ystart,yend]

    # Update the image object with our new data and extent
    im = ax.images[-1]
    cdata=mandel(xstart, xend, ystart, yend)
    im.set_data(cdata)
    crange=span(cdata)
    im.set_extent((xstart, xend, ystart, yend))
    if not (im.colorbar is None):
        fig=ax.get_figure()
        s = copy.copy( fig.canvas.toolbar._views )  
        #this store and restore is made to preserve the zooming history, 
        p = copy.copy( fig.canvas.toolbar._positions )
        im.colorbar.remove()
        plt.clim(crange)
        plt.colorbar()
        fig.canvas.toolbar._views = s
        fig.canvas.toolbar._positions = p  
    #ax.figure.canvas.draw_idle()

    
if __name__=="__main__":
    md = MandlebrotDisplay()
    Z = md(-2., 0.5, -1.25, 1.25)

    fig1, ax2 = plt.subplots()

    plt.imshow(Z, origin='lower', extent=(md.x.min(), md.x.max(), md.y.min(), md.y.max()))
    plt.colorbar()

    ax2.callbacks.connect('xlim_changed', ax_update2)
    ax2.callbacks.connect('ylim_changed', ax_update2)

    plt.show()


import matplotlib.pyplot as plt
import numpy as np
from dataIO.span import span
from pyProfile.profile import remove_profile_outliers
    
def remove_outliers2d(data,x=None,y=None,nsigma=3,fignum=None,name='',includenan=True):
    """remove outliers line by line by interpolation (along vertical lines).
    If fignum is set, plot comparison in corresponding figure."""

    ldata= data.copy()
    mask=np.apply_along_axis( remove_profile_outliers, axis=0, arr=ldata, 
        nsigma=nsigma ,includenan=includenan) #this modifies ldata and return a mask

    #these are used only to determine min and max.
    if x is None:
        x=np.arange(data.shape[1])
    if y is None:
        y=np.arange(data.shape[0])
              
    if fignum:
        #plot data before and after removal of outliers
        plt.figure(fignum)
        plt.clf()
        plt.suptitle('Effects of outliers removal - %s'%name).set_size('large')
        
        
        ax1=plt.subplot(221)
        plt.title('data')
        plt.xlabel('X (mm)')
        plt.ylabel('Y (mm)')
        axim=plt.imshow(data,interpolation='None',aspect='auto',
                        clim=span(ldata),extent=np.hstack([span(x),span(y)]))
        plt.colorbar()

        ax2=plt.subplot(222,sharex=ax1,sharey=ax1)
        plt.title('data outliers removed')
        plt.xlabel('X (mm)')
        plt.ylabel('Y (mm)')
        axim=plt.imshow(ldata,interpolation='None',aspect='auto',
                        clim=span(ldata),extent=np.hstack([span(x),span(y)]))
        plt.colorbar()

        ax3=plt.subplot(223,sharex=ax1,sharey=ax1)
        plt.title('outliers')
        outliers=ldata-data
        plt.imshow(outliers,interpolation='None',aspect='auto',
                        clim=span(ldata),extent=np.hstack([span(x),span(y)]))
        plt.colorbar()
        
        if mask.any():
            ax4=plt.subplot(224)
            plt.title('outliers height distribution')
            #plt.hist(outliers[mask].flatten(),bins=20,label='$\Delta$ correction',alpha=0.5)
            plt.hist(data[np.isfinite(data)].flatten(),bins=20,label='before',alpha=0.3,color='b')
            plt.hist(ldata[np.isfinite(ldata)].flatten(),bins=20,label='after',alpha=0.3,color='r')
            ax4b=ax4.twinx()
            ax4b.hist((data-ldata)[np.isfinite(data-ldata)].flatten(),bins=20,label='difference',alpha=0.3,color='r')
            plt.xlabel('Height (um)')
            plt.ylabel('N')
            plt.legend(loc=0)
        plt.show()
        
    return ldata

def remove_outliers2d2(data,x=None,y=None,nsigma=3,degree=1,fignum=None,name='',includenan=True):
    """remove outliers by comparing with average line profile."""
    #not working

    ldata= data.copy()
    avgprof=np.mean(ldata,axis=1)
    for i in ldata.T:
        d=(i-avgprof)
        d=d-polyfit_profile(d)  #d is the difference from average profile
        m=np.nanmean(d)
        rms=np.nanstd(d)
    mask=np.apply_along_axis( remove_profile_outliers, axis=0, arr=ldata, 
        nsigma=nsigma ,includenan=includenan) #this modifies ldata and return a mask

    #these are used only to determine min and max.
    if x is None:
        x=np.arange(data.shape[1])
    if y is None:
        y=np.arange(data.shape[0])
              
    if fignum:
        #plot data before and after removal of outliers
        plt.figure(fignum)
        plt.clf()
        plt.suptitle('Effects of outliers removal - %s'%name).set_size('large')
        
        
        ax1=plt.subplot(221)
        plt.title('data')
        plt.xlabel('X (mm)')
        plt.ylabel('Y (mm)')
        axim=plt.imshow(data,interpolation='None',aspect='auto',
                        clim=span(ldata),extent=np.hstack([span(x),span(y)]))
        plt.colorbar()

        ax2=plt.subplot(222,sharex=ax1,sharey=ax1)
        plt.title('data outliers removed')
        plt.xlabel('X (mm)')
        plt.ylabel('Y (mm)')
        axim=plt.imshow(ldata,interpolation='None',aspect='auto',
                        clim=span(ldata),extent=np.hstack([span(x),span(y)]))
        plt.colorbar()

        ax3=plt.subplot(223,sharex=ax1,sharey=ax1)
        plt.title('outliers')
        outliers=ldata-data
        plt.imshow(outliers,interpolation='None',aspect='auto',
                        clim=span(ldata),extent=np.hstack([span(x),span(y)]))
        plt.colorbar()
        
        if mask.any():
            ax4=plt.subplot(224)
            plt.title('outliers height distribution')
            #plt.hist(outliers[mask].flatten(),bins=20,label='$\Delta$ correction',alpha=0.5)
            plt.hist(data[np.isfinite(data)].flatten(),bins=20,label='before',alpha=0.3,color='b')
            plt.hist(ldata[np.isfinite(ldata)].flatten(),bins=20,label='after',alpha=0.3,color='r')
            ax4b=ax4.twinx()
            ax4b.hist((data-ldata)[np.isfinite(data-ldata)].flatten(),bins=20,label='difference',alpha=0.3,color='r')
            plt.xlabel('Height (um)')
            plt.ylabel('N')
            plt.legend(loc=0)
        plt.show()
        
    return ldata


""" vedi anche from `mpl_toolkits.axes_grid1`."""


import math
import numpy as np
from matplotlib import pyplot as plt
from dataIO.span import span
from plotting.fignumber import fignumber
from plotting.add_clickable_markers import add_clickable_markers2
from pySurf.data2D import plot_data
import pdb


def smartcb(ax=None):
    """from different places online is a way to get colorbars same height than the plot."""
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    if ax is None: ax = plt.gca()
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", "5%", pad="3%")
    return plt.colorbar(cax=cax)

def align_axis_x(ax, ax_target):
    """Make x-axis of `ax` aligned with `ax_target` in figure. Useful to align shared
    axis when colorbar messes up the alignment.
    
    from: https://stackoverflow.com/questions/46694889/matplotlib-sharex-with-colorbar-not-working
    see also:
    https://stackoverflow.com/questions/54473465/aligning-x-axis-with-sharex-using-subplots-and-colorbar-with-matplotlib
    https://stackoverflow.com/questions/44682146/align-subplot-with-colorbar
    """
    posn_old, posn_target = ax.get_position(), ax_target.get_position()
    ax.set_position([posn_target.x0, posn_old.y0, posn_target.width, posn_old.height])

def align_axis_y(ax, ax_target):
    """Make y-axis of `ax` aligned with `ax_target` in figure"""
    posn_old, posn_target = ax.get_position(), ax_target.get_position()
    ax.set_position([posn_old.x0, posn_target.y0, posn_old.width, posn_target.height])

def commonscale(fig=None):
    """set scale of all axes in a figure to the range that includes all axes data."""
    rx=[]
    ry=[]
    sax = [] #scalable axis
    if fig is None: fig=plt.gcf()
    for ax in fig.axes:
        if len(ax.images) != 0:  #dirty way to exclude e.g. colorbars
            rx.append(ax.xaxis.get_data_interval())  #under the assumption it has both x and ylim
            ry.append(ax.yaxis.get_data_interval())
            sax.append(ax)
    
    commonlim=[[min([r[0] for r in rx]),max([r[1] for r in rx])],
                [min([r[0] for r in ry]),max([r[1] for r in ry])]]
    
    for ax in sax:
        ax.set_xlim(*commonlim[0])
        ax.set_ylim(*commonlim[1])

    plt.draw()
    
    

def find_grid_size(number, smax=0, square = True, fill = False):
    #upgraded version. under development
    """given a number of plots determine the grid size that better fits all plots.
    First number returned is smaller, meaning more rows than columns.
    It is up to the user to switch for use as rows or cols. 
    Square and fill define the ideal shape for the grid.
    If fill is set returns a shape with no empty axis,
    if square is set the final shape will be as close as possible to a square,
        if not, it will be as long as it can in one dimension (meaning 
        that if smax is not set it will be a single line).
        
    for example 
    print(find_grid_size(10))  #gives 4,3
    print(find_grid_size(10,3))  #gives 4,3
    print(find_grid_size(9))  #gives 3,3
    print(find_grid_size(9,2))  #gives 5,2
    print (find_grid_size(3))  #gives 2,2
    print(find_grid_size(3,2),square=False)  #gives 2,2
    
    for i in range(11):print(i,find_grid_size(i))
    for i in range(11):print(i,find_grid_size(i,2))

    """
    
    #smax is set and you want to go for square -> square, or smax whatever is smaller
    #smax is set and you want to go for wide -> 
    #smax 0 square:1
    #smax:0 square:0
    
    if smax == 0:
        smax=number
    elif smax < 0:
        raise ValueError 
        
    #set res (array of sizes)
    if fill:
        raise NotImplementedError
    else:
        if square:
            s = int(math.sqrt(number))  # minimum possible size    
            if number == s**2:   #if exact square is ok
                res = (s, s)
            elif number <= (s+1)*s:   #needs to add one line
                res = (s+1, s)
            else:
                res = (s+1, s+1)      #if just below square, needs bigger
        else:
            maxsize = max(smax,1+number//smax)
            minsize=(number-1)//maxsize+1
            maxsize=min(number,maxsize) #adjust if 1 row
            return (minsize,maxsize)
    
    if s + 1 > smax:    
        res = ( smax,1+((number-1)//smax) )

    #if direction == 1:
    #    res = res[::-1]

    return res

from dataIO.outliers import remove_outliers

def plot_difference(p1t,p4, trim = None, dis=False):
    """plots and return difference of two Data2D objects, return difference.
    All data are plane leveled before plots, a common color scale is set after excluding outliers. Leveled difference is returned.
    If trim is other than None, plots are adjusted on valid data x and y range,
    if Trim = True empty borders are removed also from difference data."""
    
    if trim is not None:
        if trim:
            p1t=p1t.remove_nan_frame()
            p4=p4.remove_nan_frame()
        xr=span(np.array([span(d.remove_nan_frame().x) for d in [p1t,p4]]))
        yr=span(np.array([span(d.remove_nan_frame().y) for d in [p1t,p4]]))
    else:
        xr=span(np.array([span(d.x) for d in [p1t,p4]]))
        yr=span(np.array([span(d.y) for d in [p1t,p4]]))
        
    plt.clf()
    
    ax1=plt.subplot(131)
    p1t.level((1,1)).plot()
    #plt.title('PZT + IrC')
    plt.clim(*remove_outliers(p1t.level().data,nsigma=2,itmax=3,span=1))
    plt.grid()


    ax2=plt.subplot(132,sharex=ax1,sharey=ax1)
    p4.level((1,1)).plot()
    #plt.title('PZT')
    plt.clim(*remove_outliers(p4.level().data,nsigma=2,itmax=3,span=1))
    plt.grid()
    
    ax3=plt.subplot(133,sharex=ax1,sharey=ax1)
    diff=(p1t-p4).level((1,1))
    diff.name='Difference 1-2'
    diff.plot()
    plt.clim(*remove_outliers(diff.level().data,nsigma=2,itmax=3,span=1))
    plt.grid()
    plt.xlim(*xr) #this adjust all plots to common scale
    plt.ylim(*yr)   
    
    if dis:
        display(plt.gcf())
    
    return diff
    
    
def xfind_grid_size(number, smax=0):
    """given a number of plots determine the grid size that better fits all plots.
    First number returned is biggest, it is up to the user to switch 
    for use as rows or cols. 
    for example 
    print(find_grid_size(10))  #gives 4,3
    print(find_grid_size(10,3))  #gives 4,3
    print(find_grid_size(9))  #gives 3,3
    print(find_grid_size(9,2))  #gives 5,2
    
    for i in range(11):print(i,find_grid_size(i))
    for i in range(11):print(i,find_grid_size(i,2))

    """
    s = int(math.sqrt(number))  # 

    if number == s**2:
        res = (s, s)
    elif number <= (s+1)*s:
        res = (s+1, s)
    else:
        res = (s+1, s+1)
    """
    if (number-s**2) < ((s+1)**2-number):
        res = (s, s) 
    else:
        res = (s+1, s)
    """
    
    if smax > 0:
        if s + 1 > smax:
            res = (1+((number-1)//smax) , smax)

    #if direction == 1:
    #    res = res[::-1]

    return res


def subplot_grid(number,size=0,smax=0,*args,**kwargs):
    
    """
    Create a set of n=`number` subplots in a figure, automatically 
    chosing the shape of the grid. Returns figure and list of axis.
    Note that empty subplots (e.g. the 9th in a grid of 3x3 subplots when `number=8`) are not created.
    
    size: if this is provided, is used as size of the subplot grid. 
    smax: used in `find_grid_size` to limit the extension along axis for the subplot grid. See `find_grid_size` help.
    num: if integer is passed, plot on the corresponding figure. if None is passed, create a new one. TBD: pass a figure obj.
    Additional *args and **kwargs are passed to `plt.subplots`.
    
    
    2020/05/14 fixed bug on subplots removal, updated doc from:
    return a generator that plots n plots. It has advantage wrt plt.subplots of 
    not creating empty axes. The idea of making a generator is weird.
    modified to fig,axes as equivalent to plt.subplots 2018/09/06.
    Usage:
        for i,a in enumerate(subplot_grid(3)):
            a.plot(x,x**i) #or plt.plot

            also axes=[a.plot(x,x**i) for i,a in enumerate(subplot_grid(3))]"""

    gridsize = find_grid_size(number,smax) if size == 0 else size
    #fig = fignumber(fignum)

    """    if fignum==0:
        plt.clf()
    else:
        plt.figure(fignum)

    plt.clf()    """
    
    #note:
    #    #this doesn't work, if you use subplot the sharex, sharey interface is different and
    # axes references need to be passed as opposite to subplots, that uses e.g. sharex='all'
    # in that case however all axis are created at first, so I use a workaround and remove
    # excess axis.
    #fig,axes=plt.subplots(xs,ys,sharex='all',sharey='all',squeeze=True)
    #  where sharex seems to need axis reference instead of string, as a consequence also this fails:        
    """
    axes=[]
    if i==0:
        axes.append(plt.subplot(gridsize[0], gridsize[1], i+1,*args,**kwargs))
    else:
        axes.append(plt.subplot(gridsize[0], gridsize[1], i+1,*args,**kwargs))
        
    #equivalent one-liner:
    #axes = [plt.subplot(gridsize[0], gridsize[1], i+1,*args,**kwargs) for i in range(number)]
    """    
    num = kwargs.get('num',plt.gcf().number)
    fig,axes=plt.subplots(*gridsize,num=num,squeeze=True,*args,**kwargs)
    #pdb.set_trace()
    if np.size(axes) > 1:  # flag `squeeze` returns np array with removed 
        # empty dimensions if >1 element, otherwise scalar.
        # I want to have it as a list (single or multi elements) 
        axes=axes.flatten().tolist()
    else:
        axes = list(axes)
    
    # remove extra axis
    for i in np.arange(number,len(axes)): #a in axes[(np.arange(len(axes))+1)>number]: 
        a=axes.pop(i)
        a.remove()
    
    return fig,axes

''' #moved to dlist
def compare_images(datalist, x=None, y=None, fignum=None, titles=None,
                   vmin=None, vmax=None, commonscale=False, axis=0, axmax=0,
                   *args, **kwargs):
    """return a generator that plots n images in a list in subplots with shared
    zoom and axes. datalist is a list of data in format (data, x, y).
    x and y provide plot range (temporarily, ideally want to be able to 
    plot same scale).
    fignum window where to plot, if fignnum is 0 current figure is cleared,
    if None new figure is created.
    axis defines the axis with larger number of plots (default ncols >= nrows)
    axmax maximum nr of plots along shorter dim.
    """
    #modified again 2018/06/05 to accept list of data,x,y triplets. x and y are 
    # accepted as range.
    # old docstring was:
    # return a generator that plots n images in a list in subplots with shared
    # zoom and axes. datalist is a list of 2d data on same x and y.
    # fignum window where to plot, if fignnum is 0 current figure is cleared,
    # if None new figure is created.
    
    
    # this was changed a couple of times in interface, for example in passing
    # ((data,x,y),...) rather than (data1,data2),x,y
    # it can be tricky in a general way if data don't have all same size,
    # at the moment it is assumed they have.
    # in any case x and y are used only to generate the extent,
    # that is then adapted to the size of data.

    gridsize = find_grid_size(len(datalist),axmax)
    if axis == 1: gridsize=gridsize[::-1]

    fig = fignumber(fignum)

    plt.clf()
    ax = None

    # this is to set scale if not fixed
    d1std = [np.nanstd(data[0]) for data in datalist]
    std = min(d1std)

    for i, d in enumerate(datalist):
        """adjust to possible input formats"""
        data, x, y = d
        if x is None:
            x = np.arange(data.shape[1])
        if y is None:
            y = np.arange(data.shape[0])
        ax = plt.subplot(gridsize[0], gridsize[1], i+1, sharex=ax, sharey=ax)
        if titles is not None:
            print("titles is not none, it is ", titles)
            plt.title(titles[i])
        # plt.xlabel('X (mm)')
        # plt.ylabel('Y (mm)')
        s = (std if commonscale else d1std[i])
        d1mean = np.nanmean(data)
        aspect=kwargs.pop('aspect',None)
        axim = plt.imshow(data, extent=np.hstack([span(x),span(y)]),
                          interpolation='none', aspect=aspect,
                          vmin=kwargs.get('vmin', d1mean-s),
                          vmax=kwargs.get('vmax', d1mean+s),
                          *args, **kwargs)
        plt.colorbar()
        yield ax
'''

def diff_images(data1,data2,x=None,y=None,fignum=None,titles=None,vmin=None,vmax=None,
    commonscale=False, direction=0, *args, **kwargs):
    """plots two data sets with common axis and their difference. Return the three axis.
    Colorbars are formatted to be same height as plot.
    2018/06/19 use data2D routines, allowing to add parameters (e.g. stats legend).
    """

    fig=fignumber(fignum)

    plt.clf() 
    
    aspect = kwargs.pop('aspect','auto')
    #this is to set scale if not fixed
    d1std=[np.nanstd(data) for data in (data1,data2)]
    std=min(d1std)

    if x is None:
        x=np.arange(data1.shape[1])
    if y is None:
        y=np.arange(data1.shape[0])

    ax1=plt.subplot (131)
    s=(std if commonscale else d1std[0])
    d1mean=np.nanmean(data1)
    plot_data(data1,x,y,aspect=aspect,vmin=kwargs.get('vmin',d1mean-s),
        vmax=kwargs.get('vmax',d1mean+s),*args, **kwargs)    
    
    ax2=plt.subplot (132,sharex=ax1,sharey=ax1)
    s=(std if commonscale else d1std[1])
    d2mean=np.nanmean(data2)
    plot_data(data2,x,y,aspect=aspect,vmin=kwargs.get('vmin',d2mean-s),
        vmax=kwargs.get('vmax',d2mean+s),*args, **kwargs) 

    ax3=plt.subplot (133,sharex=ax1,sharey=ax1)
    plt.title('Difference (2-1)')
    diff=data2-data1
    plot_data(diff,x,y,aspect=aspect,*args, **kwargs) 

    axes=[ax1,ax2,ax3]
    if titles is not None:
        for ax,tit in zip(axes,titles):
            if tit is not None:
                ax.set_title(tit)
        
    return axes    
    
def xdiff_images(data1,data2,x=None,y=None,fignum=None,titles=None,vmin=None,vmax=None,
    commonscale=False, direction=0, *args, **kwargs):
    """plots two data sets with common axis and their difference. Return the three axis.
    Colorbars are formatted to be same height as plot.
    """

    fig=fignumber(fignum)

    plt.clf()
    ax=None
    
    aspect = kwargs.pop('aspect','auto')
    #this is to set scale if not fixed
    d1std=[np.nanstd(data) for data in (data1,data2)]
    std=min(d1std)

    if x is None:
        x=np.arange(data1.shape[1])
    if y is None:
        y=np.arange(data1.shape[0])

    ax1=plt.subplot (131,sharex=ax,sharey=ax)

    s=(std if commonscale else d1std[0])
    d1mean=np.nanmean(data1)
    axim=plt.imshow(data1,extent=np.hstack([span(x),span(y)]),
        interpolation='None',aspect=aspect,vmin=kwargs.get('vmin',d1mean-s),vmax=kwargs.get('vmax',d1mean+s), *args, **kwargs)
    plt.colorbar(fraction=0.046, pad=0.04)

    ax2=plt.subplot (132,sharex=ax,sharey=ax)

    s=(std if commonscale else d1std[1])
    d2mean=np.nanmean(data2)
    axim=plt.imshow(data2,extent=np.hstack([span(x),span(y)]),
        interpolation='None',aspect=aspect,vmin=kwargs.get('vmin',d2mean-s),vmax=kwargs.get('vmax',d2mean+s), *args, **kwargs)
    plt.colorbar(fraction=0.046, pad=0.04)

    ax3=plt.subplot (133,sharex=ax,sharey=ax)
    plt.title('Difference (2-1)')
    diff=data2-data1
    axim=plt.imshow(diff,extent=np.hstack([span(x),span(y)]),
        interpolation='None',aspect=aspect, *args, **kwargs)
    plt.colorbar(fraction=0.046, pad=0.04)

    axes=[ax1,ax2,ax3]
    if titles is not None:
        for ax,tit in zip(axes,titles):
            if tit is not None:
                ax.set_title(tit)
        
    return axes

def align_images(ax2,ax3):
    """make two subplots, allow to draw markers on them, exit at enter key and
    return markers."""
    from plotting.add_clickable_markers import add_clickable_markers2
    
    ax1,ax2=add_clickable_markers2(ax=ax2),add_clickable_markers2(ax=ax3,hold=True)
    return ax1.markers,ax2.markers

def associate_plots_axis(ax1,ax2,axis=None,on=0):
    """
    Incomplete attempt to add axis option.
    Associate axis of two different subplots `ax1`, `ax2`. 
    if `on` is None toggle, if it is 0 set off, if 1 set on. Not implemented, for now
    associate and keeps on until figure is closed.
    `axis` tells which axis are associated (0 = x, 1 =y, None = both).
    """

    if axis is None:
        axis = [0,1]  #activate both x and y-axis
    else:
        axis = [axis] # will be checked with `in`
        
    # Declare and register callbacks
    def on_xlims_change(axes):
        #print("updated xlims: ", axes.get_xlim())
        #axes.associated.set_autoscale_on(False)
        xold=axes.xold
        xl=axes.xaxis.get_view_interval()   #range of plot in data coordinates, sorted lr
        axes.xold=xl

        #print("stored value (xold), from get_view_interval (xl)=",xold,xl)
        zoom=(xl[1]-xl[0]) /(xold[1]-xold[0])
        #print("zzom",zoom)
        olim=axes.associated.get_xlim()
        #print("other axis lim (olim):",olim)
        xc=(olim[1]+olim[0])/2  #central point of other axis
        xs=(olim[1]-olim[0])/2  #half span
        #dxc=((xl[0]+xl[1])-(xold[0]+xold[1]))/2*(olim[1]-olim[0])/(xl[1]-xl[0])
        dxc=((xl[0]+xl[1])-(xold[0]+xold[1]))/2*(olim[1]-olim[0])/(xold[1]-xold[0])
        #print("offset on unzoomed data, rescaled to other axes (dxc)",dxc)
        nxl=(xc-xs*zoom+dxc,xc+xs*zoom+dxc)  #new limits
        #print(nxl)
        axes.associated.set_xlim(nxl,emit=False,auto=False)

    def on_ylims_change(axes):
        #print "updated ylims: ", axes.get_ylim()
        #axes.associated.set_autoscale_on(False)
        xold=axes.yold
        xl=axes.yaxis.get_view_interval()   #range of plot in data coordinates, sorted lr
        axes.yold=xl

        #print "stored value (yold), from get_view_interval (yl)=",xold,xl
        zoom=(xl[1]-xl[0]) /(xold[1]-xold[0])
        #print "zzom",zoom
        olim=axes.associated.get_ylim()
        #print "other axis lim (olim):",olim
        xc=(olim[1]+olim[0])/2  #central point of other axis
        xs=(olim[1]-olim[0])/2  #half span
        dxc=((xl[0]+xl[1])-(xold[0]+xold[1]))/2*(olim[1]-olim[0])/(xold[1]-xold[0])
        #print "offset on unzoomed data, rescaled to other axes (dxc)",dxc
        #nxl=(xc+dxc,xc)  #newlimits
        nxl=(xc-xs*zoom+dxc,xc+xs*zoom+dxc)
        axes.associated.set_ylim(nxl,emit=False,auto=False)

    def ondraw(event):
        # print 'ondraw', event
        # these ax.limits can be stored and reused as-is for set_xlim/set_ylim later
        #print event -> it is matplotlib.backend_bases.DrawEvent object (?)
        ax1.xold=ax1.get_xlim()
        ax2.xold=ax2.get_xlim()
        ax1.yold=ax1.get_ylim()
        ax2.yold=ax2.get_ylim()
        #print "on draw axis: axes xlims", ax1.xold, ax2.xold


    fig=ax1.figure
    cid1 = fig.canvas.mpl_connect('draw_event', ondraw)
    if ax2.figure != fig:
        cid2 = ax2.figure.canvas.mpl_connect('draw_event', ondraw)
    
    setattr(ax2,'associated',ax1)
    setattr(ax2,'associated',ax1)
    
    if 0 in axis:
        setattr(ax1,'xold',ax1.get_xlim())
        setattr(ax2,'xold',ax2.get_xlim())
        ax1.callbacks.connect('xlim_changed', on_xlims_change)
        ax2.callbacks.connect('xlim_changed', on_xlims_change)        
        
    if 1 in axis:
        setattr(ax1,'yold',ax1.get_ylim())
        setattr(ax2,'yold',ax2.get_ylim())
        ax1.callbacks.connect('ylim_changed', on_ylims_change)
        ax2.callbacks.connect('ylim_changed', on_ylims_change)


def associate_plots(ax1,ax2,axis=None,on=0):
    """
    Associate axis of two different subplots `ax1`, `ax2`.
    This is a different mechanism than shared axis, in that here zoom and pans are kept synchronized, but axis are independent. If the two plots have same scales on both axis, it works like shared axis.
    
    Run `test_associate`, `test_associate_zoom` for examples.
    
    TODO: toggle association: note that relative alignment of centers is maintained, to be able to manually align center a toggle mechanism must be implemented:
    if `on` is None toggle, if it is 0 set off, if 1 set on. Not implemented, for now
    associate and keeps on until figure is closed.
    TODO: being able to associate only on x or y `axis` tells which axis are associated (0 = x, 1 =y, None = both). See attempt in `associate_plots_axis`
    TODO: when home is clicked, plots are messed up, need to click twice.
    
    """

    if axis is None:
        axis = [0,1] #activate both x and y-axis
        
    # Declare and register callbacks
    def on_xlims_change(axes):
        print("updated xlims: ", axes.get_xlim())
        #axes.associated.set_autoscale_on(False)
        xold=axes.xold
        xl=axes.xaxis.get_view_interval()   #range of plot in data coordinates, sorted lr
        axes.xold=xl

        print("stored value (xold), from get_view_interval (xl)=",xold,xl)
        zoom=(xl[1]-xl[0]) /(xold[1]-xold[0])
        print("zzom",zoom)
        olim=axes.associated.get_xlim()
        print("other axis lim (olim):",olim)
        xc=(olim[1]+olim[0])/2  #central point of other axis
        xs=(olim[1]-olim[0])/2  #half span
        #dxc=((xl[0]+xl[1])-(xold[0]+xold[1]))/2*(olim[1]-olim[0])/(xl[1]-xl[0])
        dxc=((xl[0]+xl[1])-(xold[0]+xold[1]))/2*(olim[1]-olim[0])/(xold[1]-xold[0])
        print("offset on unzoomed data, rescaled to other axes (dxc)\n\n",dxc)
        nxl=(xc-xs*zoom+dxc,xc+xs*zoom+dxc)  #new limits
        #print(nxl)
        axes.associated.set_xlim(nxl,emit=False,auto=False)

    def on_ylims_change(axes):
        print ("updated ylims: ", axes.get_ylim())
        #axes.associated.set_autoscale_on(False)
        xold=axes.yold
        xl=axes.yaxis.get_view_interval()   #range of plot in data coordinates, sorted lr
        axes.yold=xl

        print ("stored value (yold), from get_view_interval (yl)=",xold,xl)
        zoom=(xl[1]-xl[0]) /(xold[1]-xold[0])
        print ("zzom",zoom)
        olim=axes.associated.get_ylim()
        print ("other axis lim (olim):",olim)
        xc=(olim[1]+olim[0])/2  #central point of other axis
        xs=(olim[1]-olim[0])/2  #half span
        dxc=((xl[0]+xl[1])-(xold[0]+xold[1]))/2*(olim[1]-olim[0])/(xold[1]-xold[0])
        print ("offset on unzoomed data, rescaled to other axes (dxc)\n\n",dxc)
        #nxl=(xc+dxc,xc)  #newlimits
        nxl=(xc-xs*zoom+dxc,xc+xs*zoom+dxc)
        axes.associated.set_ylim(nxl,emit=False,auto=False)

    def ondraw(event):
        # print 'ondraw', event
        # these ax.limits can be stored and reused as-is for set_xlim/set_ylim later
        #print event -> it is matplotlib.backend_bases.DrawEvent object (?)
        ax1.xold=ax1.get_xlim()
        ax2.xold=ax2.get_xlim()
        ax1.yold=ax1.get_ylim()
        ax2.yold=ax2.get_ylim()
        #print "on draw axis: axes xlims", ax1.xold, ax2.xold


    fig=ax1.figure
    cid1 = fig.canvas.mpl_connect('draw_event', ondraw)
    if ax2.figure != fig:
        cid2 = ax2.figure.canvas.mpl_connect('draw_event', ondraw)

    setattr(ax1,'associated',ax2)
    setattr(ax1,'xold',ax1.get_xlim())
    setattr(ax1,'yold',ax1.get_ylim())

    setattr(ax2,'associated',ax1)
    setattr(ax2,'xold',ax2.get_xlim())
    setattr(ax2,'yold',ax2.get_ylim())

    ax1.callbacks.connect('xlim_changed', on_xlims_change)
    ax1.callbacks.connect('ylim_changed', on_ylims_change)
    ax2.callbacks.connect('xlim_changed', on_xlims_change)
    ax2.callbacks.connect('ylim_changed', on_ylims_change)

def associate_plots_mw(ax1,ax2):
    """if on is None toggle, if it is 0 set off, if 1 set on. Not implemented, for now
    associate and keeps on until figure is closed.
    """

    # Declare and register callbacks
    def on_xlims_change(axes):

        # print "updated xlims: ", axes.get_xlim()
        # axes.associated.set_autoscale_on(False)
        xold = axes.xold
        xl = axes.xaxis.get_view_interval()   # range of plot in data coordinates, sorted lr
        axes.xold = xl

        zoom = (xl[1]-xl[0]) / (xold[1]-xold[0])
        olim = axes.associated.get_xlim()
        print("---------x limit change-------")
        print("on ax1\n" if axes == ax1 else "on ax2\n")
        print("old: (%5.3f,%5.3f)" % tuple(xold),
              "new:(%5.3f,%5.3f)" % tuple(xl),
              " zoom:%5.3f" % zoom)
        xc = (olim[1]+olim[0])/2  # central point of other axis
        xs = (olim[1]-olim[0])/2  # half span
        dxc = ((xl[0]+xl[1])-(xold[0]+xold[1]))/2*(olim[1]-olim[0])/(xold[1]-xold[0])
        # print "offset on unzoomed data, rescaled to other axes (dxc)",dxc
        nxl = (xc-xs*zoom+dxc, xc+xs*zoom+dxc)  # new limits
        print("new range other axes:(%5.3f,%5.3f)\n" % tuple(nxl))
        axes.associated.set_xlim(nxl, emit=False, auto=False)


    def ondraw(event):
        print('------ ondraw --------')
        # these ax.limits can be stored and reused as-is for set_xlim/set_ylim later
        ax1.xold = ax1.get_xlim()
        ax2.xold = ax2.get_xlim()
        print("axes xlims: (%5.3f,%5.3f)" % (ax1.xold),
              ",(%5.3f,%5.3f)" % (ax2.xold), "\n")

    fig = ax1.figure
    cid1 = fig.canvas.mpl_connect('draw_event', ondraw)
    if ax2.figure != fig:
        cid2 = ax2.figure.canvas.mpl_connect('draw_event', ondraw)

    setattr(ax1, 'associated', ax2)
    setattr(ax1, 'xold', ax1.get_xlim())
    setattr(ax2, 'associated', ax1)
    setattr(ax2, 'xold', ax2.get_xlim())

    ax1.callbacks.connect('xlim_changed', on_xlims_change)
    ax2.callbacks.connect('xlim_changed', on_xlims_change)

def associate_plots_muc(ax1,ax2):
    """just print the changes of axes.
    """

    def on_xlims_change(axes):
        xold=axes.xold
        xl=axes.xaxis.get_view_interval()   #range of plot in data coordinates, sorted lr
        axes.xold=xl
        olim=axes.associated.get_xlim()
        print("---------x limit change-------")
        print("on ax1\n" if axes==ax1 else "on ax2\n")
        print("old: (%5.3f,%5.3f)"%tuple(xold), "new:(%5.3f,%5.3f)"%tuple(xl))
        axes.associated.set_xlim((5,8),emit=False,auto=False)

    def ondraw(event):
        print('------ ondraw --------')
        # these ax.limits can be stored and reused as-is for set_xlim/set_ylim later
        ax1.xold=ax1.get_xlim()
        ax2.xold=ax2.get_xlim()
        print("axes xlims: (%5.3f,%5.3f)"%(ax1.xold),",(%5.3f,%5.3f)"%(ax2.xold),"\n")

    cid1 = ax1.figure.canvas.mpl_connect('draw_event', ondraw)
    cid2 = ax2.figure.canvas.mpl_connect('draw_event', ondraw)

    setattr(ax1,'associated',ax2)
    setattr(ax1,'xold',ax1.get_xlim())
    setattr(ax2,'associated',ax1)
    setattr(ax2,'xold',ax2.get_xlim())

    ax1.callbacks.connect('xlim_changed', on_xlims_change)
    ax2.callbacks.connect('xlim_changed', on_xlims_change)

def associate_plots_tb2(ax1,ax2):
    """attempt to use an alternative mechanism for adjusting plot scales
        based on toolbar history. potentially lighter and more robust since someone
        else took care of details. need to keep under control side effects,
        for example must work on rescaling from code.
    """
    from matplotlib.backend_bases import NavigationToolbar2

    def ondraw(event):
        print('------ ondraw --------')
        # these ax.limits can be stored and reused as-is for set_xlim/set_ylim later
        ax1.xold=ax1.get_xlim()
        ax2.xold=ax2.get_xlim()
        if not ax1.figure.canvas.toolbar._views.home() is None:

            print(ax1.figure.canvas.toolbar._views.home())

            print(np.array([aa[:2] for aa in ax1.figure.canvas.toolbar._views.home()]))
            print(np.array(ax1.xold))
            if (np.array([aa[:2] for aa in ax1.figure.canvas.toolbar._views.home()])==np.array(ax1.xold)).all():
                print("home?!")
        print("axes xlims: (%5.3f,%5.3f)"%(ax1.xold),",(%5.3f,%5.3f)"%(ax2.xold),"\n")

    cid1 = ax1.figure.canvas.mpl_connect('draw_event', ondraw)
    cid2 = ax2.figure.canvas.mpl_connect('draw_event', ondraw)

    setattr(ax1,'associated',ax2)
    setattr(ax1,'xold',ax1.get_xlim())
    setattr(ax2,'associated',ax1)
    setattr(ax2,'xold',ax2.get_xlim())

    ax1.callbacks.connect('xlim_changed', on_xlims_change)
    ax2.callbacks.connect('xlim_changed', on_xlims_change)

def set_home():
    from matplotlib.backend_bases import NavigationToolbar2
    home = NavigationToolbar2.home

    def new_home(self, *args, **kwargs):
        print('NEW HOME!')
        #print "method:",home

        home(self, *args, **kwargs)

    NavigationToolbar2.home = new_home
    return home

def restore_home(h):
    from matplotlib.backend_bases import NavigationToolbar2
    NavigationToolbar2.home=h

def test_associate():
    a=np.random.random(20).reshape((5,4))
    b=-a
    ax1=plt.subplot(121)
    plt.imshow(a,interpolation='none')
    ax2=plt.subplot(122)
    plt.imshow(b,interpolation='none')#,extent=[0,10,-1,1],aspect='equal')

    associate_plots(ax1,ax2,on=1)
    plt.show()
    return ax1,ax2

def test_associate_zoom():
    np.set_printoptions(3)
    a=np.random.random(20).reshape((5,4))
    b=a
    ax1=plt.subplot(121)
    plt.imshow(a,interpolation='none',extent=[0,20,-2,2],aspect='equal')
    ax2=plt.subplot(122)
    plt.imshow(b,interpolation='none',extent=[0,10,-1,1],aspect='equal')

    #associate_plots_muc(ax1,ax2)
    #associate_plots_tb2(ax1,ax2)
    associate_plots(ax1,ax2)
    associate_plots_mw(ax1,ax2)
    plt.show()
    return ax1,ax2

def test_subplot_grid():
    """test/example for subplot_grid"""
    
    plt.ion()
    
    #from plotting.multiplots import subplot_grid
    x=np.array([1,2,3])
    plt.figure(1)
    plt.clf()
    plt.suptitle('subplot_grid(3)')
    for i,a in enumerate(subplot_grid(3)[1]):
        a.set_title('x**%i'%i)
        a.plot(x, x**i)
    plt.tight_layout()

    plt.figure(2)
    plt.clf()
    for i,a in enumerate(subplot_grid(3)[1]):
        plt.plot(x,x**i)

    plt.figure(3)
    plt.clf()
    axes=[a.plot(x,x**i) for i,a in enumerate(subplot_grid(3)[1])]

    plt.figure(5)
    f,s=subplot_grid(8)
    for i,ss in enumerate(s):
        ss.plot(np.array([1,2,3])+i,[3,4,5])


if __name__=='__main__':
    for i in range(10,20):
        print(i, find_grid_size(i))


import numpy as np

def plane_fit (x, y, z):
    """
    #modified by kov
    #x,y,z 3 vectors with coordinates of points (same number of elements).
    #return value [A,B,C] of plane Ax + By + C = z
    #planesurf is a vector with z of plane points

    #compute the average surface, calculate statistical indicator
    # z can be nan, if there are nan in x and y it is likely not to woork.

      # M. Katz 1/26/04
    # IDL function to perform a least-squares fit a plane, based on
    # Ax + By + C = z
    #
    # ABC = plane_fit(x, y, z, error=error)
    """
    #check length of vectors.
    if len(y) != len(x):
        raise ValueError("wrong number of elements for x and y.")
    if len(z) != len(x):
        raise ValueError("wrong number of elements for z.")
    
    gi=~np.isnan(z)
    x=x[gi]
    y=y[gi]
    z=z[gi]
    
    tx2 = np.sum(x**2)
    ty2 = np.sum(y**2)
    txy = np.sum(x*y)
    tx = np.sum(x)
    ty = np.sum(y)
    N = len(z)

    a = np.matrix([[tx2, txy, tx],
        [txy, ty2, ty],
        [tx, ty, N ]])

    b = np.array([np.nansum(z*x), np.nansum(z*y), np.nansum(z)])
    out = np.array(a.I * b[:,None])
    #plane=(out[0]*x+out[1]*y+out[2])

    return out

if __name__=='__main__':
    #plane parameters
    a=555.5
    b=-3
    c=-15

    #grid
    npx=100
    npy=100

    xp=np.arange(npx)
    yp=np.arange(npy)
    x,y=list(map(np.ndarray.flatten, np.meshgrid(xp,yp)))
    z=x*a+y*b+c

    print("starting value:",a,b,c)
    print(" fit results: ",plane_fit(x,y,z))

  


#from pool/pipeline 
import re
from itertools import chain
import os
"""extract from a list of notebooks 
a set of variables as in configuration files."""

prefixes=['gfile','cfile','markers1','markers2','gscale','mscale','rect1','c2file']
#notebooks=['OP1S16_analysis.ipynb']
notebooks=[l.strip() for l in open('listdir.dat','r').readlines()]

outconf=[]
for nb in notebooks:
	key=os.path.splitext(os.path.basename(nb))[0]
	l=[ll for ll in open(nb,'r').readlines()]	
	a=[re.findall(r'"(.*?)"', ll.strip()) for ll in l]
	b=list(chain.from_iterable([aa for aa in a if len(aa)>0]))
	b=[ll[:-2] if ll.endswith('\\n') else ll for ll in b]
	#a=[re.search(r'\"(.+?)\"',ll) for ll in l]
	#b=[aa.string[aa.start():aa.end()] for aa in a if aa is not None]
	good=[]
	for prefix in prefixes:
		c=[bb.strip() for bb in b if bb.strip().startswith(prefix)]
		good.extend(c)
	outconf.append("["+key+"]")
	outconf.append("\n".join(good))
	outconf.append("")

outconf='\n'.join(outconf)
open('outconf.ini','w').write(outconf)
print(outconf)
	


import matplotlib.pyplot as plt

def fignumber(fignum=None,*args,**kwargs):
    """ Set and return a figure given a figure number (0 for current, None for new).
    
    If `fignum` is None create a new figure, if 0 uses current figure after cleaning it, if other number, use as number of the figure. Return a `plt.figure` object.
    
    ex. fig=fignumber(0) # clear and use current figure. 

    Offers a way to use a figure number with consistent rules including opening a new window or clearing the old one (?)."""
    
    if fignum == 0:
        fig=plt.gcf()
    else:
        fig=plt.figure(fignum,*args,**kwargs)
    plt.clf()
    
    return fig
    


# -*- coding: utf-8 -*-

#from pykat.optics.romhom import makeWeightsNew
from scipy.interpolate import interp2d, interp1d
from scipy.optimize import minimize
#from pykat.maths.zernike import *        
#from pykat.exceptions import BasePyKatException
from copy import deepcopy

import numpy as np
import math
import pickle


'''
"""
------------------------------------------------------
Utility functions for handling mirror surface
maps. Some functions based on earlier version
in Matlab (http://www.gwoptics.org/simtools/)
Work in progress, currently these functions are
untested!

http://www.gwoptics.org/pykat/
------------------------------------------------------
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from pykat.optics.romhom import makeWeightsNew
from scipy.interpolate import interp2d, interp1d
from scipy.optimize import minimize
from pykat.maths.zernike import *        
from pykat.exceptions import BasePyKatException
from copy import deepcopy

import numpy as np
import math
import pickle

class MirrorROQWeights:
    
    def __init__(self, rFront, rBack, tFront, tBack):
        self.rFront = rFront
        self.rBack = rBack
        self.tFront = tFront
        self.tBack = tBack
    
    def writeToFile(self, romfilename):
        with open(romfilename + ".rom", "w+") as f:
            if self.rFront is not None: self.rFront.writeToFile(f=f)
            if self.rBack  is not None: self.rBack.writeToFile(f=f)
            if self.tFront is not None: self.tFront.writeToFile(f=f)
            if self.tBack  is not None: self.tBack.writeToFile(f=f)
                    
class surfacemap(object):
    def __init__(self, name, maptype, size, center, step_size, scaling, data=None,
                 notNan=None, Rc=None, zOffset=None, xyOffset=None):
        
        self.name = name
        self.type = maptype
        self.center = center
        self.step_size = step_size
        self.scaling = scaling
        self.notNan = notNan
        self.Rc = Rc
        self.zOffset = zOffset
        self.__interp = None
        self._zernikeRemoved = {}
        self._betaRemoved = None
        self._xyOffset = xyOffset
		
        if data is None:
            self.data = np.zeros(size)
        else:
            self.data = data
            
        if notNan is None:
            self.notNan = np.ones(size)
        else:
            self.notNan = notNan

        self._rom_weights = None
        
    def write_map(self, filename):
        with open(filename,'w') as mapfile:
            
            mapfile.write("% Surface map\n")
            mapfile.write("% Name: {0}\n".format(self.name))
            mapfile.write("% Type: {0}\n".format(self.type))
            mapfile.write("% Size: {0} {1}\n".format(self.data.shape[0], self.data.shape[1]))
            mapfile.write("% Optical center (x,y): {0} {1}\n".format(self.center[0], self.center[1]))
            mapfile.write("% Step size (x,y): {0} {1}\n".format(self.step_size[0], self.step_size[1]))
            mapfile.write("% Scaling: {0}\n".format(float(self.scaling)))
            mapfile.write("\n\n")
            
            for i in range(0, self.data.shape[0]):
                for j in range(0, self.data.shape[1]):
                    mapfile.write("%.15g " % self.data[i,j])
                mapfile.write("\n")


    @property
    def betaRemoved(self):
        return self._betaRemoved
    
    @betaRemoved.setter
    def betaRemoved(self, v):
        self._betaRemoved = v
        
    @property
    def zernikeRemoved(self):
        return self._zernikeRemoved
    
    @zernikeRemoved.setter
    def zernikeRemoved(self, v):
        """
        v = tuple(m,n,amplitude)
        """
        self._zernikeRemoved["%i%i" % (v[0],v[1])] = v

    @property
    def data(self):
        return self.__data
    
    @data.setter
    def data(self, value):
        self.__data = value
        self.__interp = None
    
    @property
    def center(self):
        return self.__center
    
    @center.setter
    def center(self, value):
        self.__center = value
        self.__interp = None
    
    @property
    def step_size(self):
        return self.__step_size
    
    @step_size.setter
    def step_size(self, value):
        self.__step_size = value
        self.__interp = None

    @property
    def scaling(self):
        return self.__scaling
    
    @scaling.setter
    def scaling(self, value):
        self.__scaling = value
        self.__interp = None

    # CHANGED! I swapped: self.data.shape[0]  <----> self.data.shape[1]. Because
    # the rows/columns in the data matrix are supposed to be y/x-values(?).
    # This may mess things up in other methods though... I fixed the plot-method.
    # /DT
    @property
    def x(self):
        return self.step_size[0] * (np.array(range(0, self.data.shape[1])) - self.center[0])
    
    @property
    def y(self):
        return self.step_size[1] * (np.array(range(0, self.data.shape[0]))- self.center[1])
        
    # CHANGED! Since everything else (step_size, center) are given as (x,y), and not
    # as (row, column), I changed this to the same format. /DT
    @property
    def size(self):
        return self.data.shape[::-1]
        
    @property
    def offset(self):
        return np.array(self.step_size)*(np.array(self.center) - (np.array(self.size)-1)/2.0)
    
    @property
    def ROMWeights(self):
        return self._rom_weights
    
    def z_xy(self, x=None, y=None, wavelength=1064e-9, direction="reflection_front", nr1=1.0, nr2=1.0):
        """
        For this given map the field perturbation is computed. This data
        is used in computing the coupling coefficient. It returns a grid
        of complex values representing the change in amplitude or phase
        of the field.
        
            x, y      : Points to interpolate at, 'None' for no interpolation.
            
            wavelength: Wavelength of light in vacuum [m]
            
            direction : Sets which distortion to return, as beams travelling
                        in different directions will see different distortions.
                        Options are:
                                "reflection_front"
                                "transmission_front" (front to back)
                                "transmission_back" (back to front)
                                "reflection_back"
                                
            nr1       : refractive index on front side
            
            nr2       : refractive index on back side
            
        """
        
        assert(nr1 >= 1)
        assert(nr2 >= 1)
        
        if x is None and y is None:
            data = self.scaling * self.data
        else:
            if self.__interp is None:
                self.__interp = interp2d(self.x, self.y, self.data * self.scaling)
                
            data = self.__interp(x, y)
        
        if direction == "reflection_front" or direction == "reflection_back":
            if "phase" in self.type:
                k = math.pi * 2 / wavelength
                
                if direction == "reflection_front":
                    return np.exp(-2j * nr1 * k * data)
                else:
                    return np.exp(2j * nr2 * k * data[:,::-1])
                
            elif "absorption" in self.type:
                if direction == "reflection_front":
                    return np.sqrt(1.0 - data)
                else:
                    return np.sqrt(1.0 - data[:, ::-1])
            elif "surface_motion" in self.type:
                if direction == "reflection_front":
                    return data
                else:
                    return data[:, ::-1]
            else:
                raise BasePyKatException("Map type needs handling")
                
        elif direction == "transmission_front" or direction == "transmission_back":
            if "phase" in self.type:
                k = math.pi * 2 / wavelength
                
                if direction == "transmission_front":
                    return np.exp((nr1-nr2) * k * data)
                else:
                    return np.exp((nr2-nr1) * k * data[:, ::-1])
                
            elif "absorption" in self.type:
                if direction == "transmission_front":
                    return np.sqrt(1.0 - data)
                else:
                    return np.sqrt(1.0 - data[:, ::-1])
                    
            elif "surface_motion" in self.type:
                return np.ones(data.shape)
            else:
                raise BasePyKatException("Map type needs handling")
                
        else:
            raise ValueError("Direction not valid")
        

    
    def generateROMWeights(self, EIxFilename, EIyFilename=None, nr1=1.0, nr2=1.0, verbose=False, interpolate=False, newtonCotesOrder=8):
        
        if interpolate == True:
            # Use EI nodes to interpolate if we
            with open(EIxFilename, 'rb') as f:
                EIx = pickle.load(f)

            if EIyFilename is None:
                EIy = EIx
            else:
                with open(EIyFilename, 'rb') as f:
                    EIy = pickle.load(f)

            x = EIx.x
            x.sort()
            nx = np.unique(np.hstack((x, -x[::-1])))
        
            y = EIy.x
            y.sort()
            ny = np.unique(np.hstack((y, -y[::-1])))
            
            self.interpolate(nx, ny)
        
        w_refl_front, w_refl_back, w_tran_front, w_tran_back = (None, None, None, None)
        
        if "reflection" in self.type or "both" in self.type:
            w_refl_front = makeWeightsNew(self, EIxFilename, EIyFilename,
                                      verbose=verbose, newtonCotesOrderMapWeight=newtonCotesOrder,
                                      direction="reflection_front")
            
            w_refl_front.nr1 = nr1
            w_refl_front.nr2 = nr2
            
            w_refl_back = makeWeightsNew(self, EIxFilename, EIyFilename,
                                      verbose=verbose, newtonCotesOrderMapWeight=newtonCotesOrder,
                                      direction="reflection_back")
            
            w_refl_back.nr1 = nr1
            w_refl_back.nr2 = nr2

        if "transmission" in self.type or "both" in self.type:                                      
            w_tran_front = makeWeightsNew(self, EIxFilename, EIyFilename,
                                      verbose=verbose, newtonCotesOrderMapWeight=newtonCotesOrder,
                                      direction="transmission_front")

            w_refl_front.nr1 = nr1
            w_refl_front.nr2 = nr2
                                            
            w_tran_back  = makeWeightsNew(self, EIxFilename, EIyFilename,
                                      verbose=verbose, newtonCotesOrderMapWeight=newtonCotesOrder,
                                      direction="transmission_back")
            
            w_refl_back.nr1 = nr1
            w_refl_back.nr2 = nr2
            
        self._rom_weights = MirrorROQWeights(w_refl_front, w_refl_back, w_tran_front, w_tran_back)
        
        return self._rom_weights
            
    def interpolate(self, nx, ny, **kwargs):
        """
        Interpolates the map for some new x and y values.
        
        Uses scipy.interpolate.interp2d and any keywords arguments are
        passed on to it, thus settings like interpolation type and
        fill values can be set.
        
        The range of nx and ny must contain the value zero so that the
        center point of the map can be set.
        """

        D = interp2d(self.x, self.y, self.data, **kwargs)
        
        data = D(nx-self.offset[0], ny-self.offset[1])
        
        Dx = interp1d(nx, np.arange(0,len(nx)))
        Dy = interp1d(ny, np.arange(0,len(ny)))
        
        self.center = (Dx(0), Dy(0))
        self.step_size = (nx[1]-nx[0], ny[1]-ny[0])
        self.data = data

    # xlim and ylim given in centimeters
    def plot(self, show=True, clabel=None, xlim=None, ylim=None):
        import pylab
        
        if xlim is not None:
            # Sorts out the x-values within xlim
            _x = np.logical_and(self.x<=max(xlim)/100.0, self.x>=min(xlim)/100.0)
            xmin = np.min(np.where(_x == True))
            xmax = np.max(np.where(_x == True))
        else:
            # Uses the whole available x-range
            xmin = 0
            xmax = len(self.x)-1
            xlim = [self.x.min()*100, self.x.max()*100]
    
        if ylim is not None:
            # Sorts out the y-values within ylim
            _y = np.logical_and(self.y<=max(ylim)/100.0, self.y>=min(ylim)/100.0)
            ymin = np.min(np.where(_y == True))
            ymax = np.max(np.where(_y == True))
        else:
            # Uses the whole available y-range
            ymin = 0
            ymax = len(self.y)-1
            ylim = [self.y.min()*100, self.y.max()*100]
            
        # CHANGED! y corresponds to rows and x to columns, so this should be
        # correct now. Switch the commented/uncommented things to revert. /DT
        # ------------------------------------------------------
        # min and max of z-values
        zmin = self.data[ymin:ymax,xmin:xmax].min()
        zmax = self.data[ymin:ymax,xmin:xmax].max()
        # zmin = self.data[xmin:xmax,ymin:ymax].min()
        # zmax = self.data[xmin:xmax,ymin:ymax].max()
        # ------------------------------------------------------
        
        # 100 factor for scaling to cm
        xRange = 100*self.x
        yRange = 100*self.y
        # xRange, yRange = np.meshgrid(xRange,yRange)
        
        fig = pylab.figure()

        # Important to remember here is that xRange corresponds to column and
        # yRange to row indices of the matrix self.data.
        pcm = pylab.pcolormesh(xRange, yRange, self.data)
        pcm.set_rasterized(True)
        
        pylab.xlabel('x [cm]')
        pylab.ylabel('y [cm]')

        if xlim is not None: pylab.xlim(xlim)
        if ylim is not None: pylab.ylim(ylim)
            
        pylab.title('Surface map {0}, type {1}'.format(self.name, self.type))

        cbar = pylab.colorbar()
        #cbar.set_clim(zmin, zmax)
        
        if clabel is not None:
            cbar.set_label(clabel)
    
        if show:
            pylab.show()
        
        return fig

    def rescaleData(self,scaling):
        """
        Rescaling data. Assumes that the current map is scaled correctly, i.e.,
        surfacemap.scaling*surfacemap.data would give the data in meters.

        Input: scaling
        scaling  - float number that representing the new scaling. E.g.,
                   scaling = 1.0e-9 means that surfacemap.data is converted to
                   being given in nano meters.
        """
        if isinstance(scaling, float):
            self.data[self.notNan] = self.data[self.notNan]*self.scaling/scaling
            self.scaling = scaling
        else:
            print('Error: Scaling factor must be of type float, {:s} found'.
                  format(type(scaling)))
            

    def rms(self, w=None):
        if w is None:
            return math.sqrt((self.data[self.notNan]**2).sum())/self.notNan.sum()
        else:
            R = self.find_radius(unit='meters')
            if w>=R:
                return math.sqrt((self.data[self.notNan]**2).sum())/self.notNan.sum()
            else:
                rho = self.createPolarGrid()[0]
                inside = np.zeros(self.data.shape,dtype=bool)
                tmp = rho<w
                inside[tmp] = self.notNan[tmp]
                return math.sqrt((self.data[inside]**2).sum())/inside.sum()
            
    def avg(self, w=None):
        if w is None:
            tot = self.data[self.notNan].sum()
            sgn = np.sign(tot)
            return sgn*math.sqrt(sgn*tot)/self.notNan.sum()
        else:
            R = self.find_radius(unit='meters')
            if w>=R:
                tot = self.data[self.notNan].sum()
                sgn = np.sign(tot)
                return sgn*math.sqrt(sgn*tot)/self.notNan.sum()
            else:
                rho = self.createPolarGrid()[0]
                inside = np.zeros(self.data.shape,dtype=bool)
                tmp = rho<w
                inside[tmp] = self.notNan[tmp]
                tot = self.data[inside].sum()
                sgn = np.sign(tot)
                return sgn*math.sqrt(sgn*tot)/inside.sum()
            
            
    def find_radius(self, method = 'max', unit='points'):
        """
        Estimates the radius of the mirror in the xy-plane. 
        method   - 'max' gives maximal distance from centre to the edge.
                   'xy' returns the distance from centre to edge along x- and y-directions.
                   'area' calculates the area and uses this to estimate the mean radius.
        unit     - 'points' gives radius in data points.
                   'meters' gives radius in meters.

        Based on Simtools function 'FT_find_map_radius.m' by Charlotte Bond.
        """

        # Row and column indices of non-NaN
        rIndx, cIndx = self.notNan.nonzero()
        if unit == 'meters' or unit == 'metres' or unit=='Meters' or unit=='Metres':
            # Offsets so that (0,0) is in the center.
            x = self.step_size[0]*(cIndx - self.center[0])
            y = self.step_size[1]*(rIndx - self.center[1])
        else:
            # Offsets so that (0,0) is in the center.
            x = cIndx - self.center[0]
            y = rIndx - self.center[1]
            
        # Maximum distance from center to the edge of the mirror.
        if method=='max' or method=='Max' or method=='MAX':
            r = math.sqrt((x**2 + y**2).max())
        # x and y radii
        elif method=='xy' or method=='XY' or method == 'yx' or method == 'YX':
            r = []
            r.append(abs(x).max()+0.5)
            r.append(abs(y).max()+0.5)
        # Mean radius by dividing the area by pi. Should change this in case
        # x and y step sizes are different.
        elif method=='area' or method=='Area' or method=='AREA':
            if unit == 'meters' or unit == 'metres' or unit=='Meters' or unit=='Metres':
                r = step_size[0]*math.sqrt(len(cIndx)/math.pi)
            else:
                r = math.sqrt(len(cIndx)/math.pi)
        elif method=='min':
            # Arrays of row and column indices where data is NaN.
            r,c = np.where(self.notNan == False)
            # Sets the map center to index [0,0]
            x = c-self.center[0]
            y = r-self.center[1]
            if unit == 'meters' or unit == 'metres' or unit=='Meters' or unit=='Metres':
                x = x*self.step_size[0]
                y = y*self.step_size[1]
            # Minimum radius squared
            r = math.sqrt((x**2+y**2).min())
        return r


    def zernikeConvol(self,n,m=None):
        """
        Performs a convolution with the map surface and Zernike polynomials.

        Input: n, m
        n  -  If m (see below) is set to None, n is the maximum order of Zernike
              polynomials that will be convolved with the map. If m is an integer
              or a list, only the exact order n will be convolved.
        m  -  = None. All Zernike polynomials up to and equal to order n will
              be convolved.
              = value [int]. The Zernike polynomial characterised by (n,m) will
              be convolved witht the map.
              = list of integers. The Zernike polynomials of order n with m-values
              in m will be convolved with the map.
             
        Returns: A
        A  -  If m is None:
                List with lists of amplitude coefficients. E.g. A[n] is a list of
                amplitudes for the n:th order with m ranging from -n to n, i.e.,
                A[n][0] is the amplitude of the Zernike polynomial Z(m,n) = Z(-n,n).
              If m is list or int:
                Returns list with with Zernike amplitudes ordered in the same way as
                in the input list m. If m == integer, there is only one value in the
                list A.
        
        Based on the simtool function 'FT_zernike_map_convolution.m' by Charlotte
        Bond.
        """
        
        mVals = []
        if m is None:
            nVals = range(0,n+1)
            for k in nVals:
                mVals.append(range(-k,k+1,2))
        elif isinstance(m,list):
            nVals = range(n,n+1)
            mVals.append(m)
        elif isinstance(m,int):
            nVals = range(n,n+1)
            mVals.append(range(m,m+1))
            
        # Amplitudes
        A = []
        # Radius
        R = self.find_radius(unit='meters')
        
        # Grid for for Zernike-polynomials (same size as map).
        rho, phi = self.createPolarGrid()
        rho = rho/R
        # Loops through all n-values up to n_max
        for k in range(len(nVals)):
            n = nVals[k]
            # Loops through all possible m-values for each n.
            tmp = []
            for m in mVals[k]:
                # (n,m)-Zernike polynomial for this grid.
                Z = zernike(m, n, rho, phi)
                # Z = znm(n, m, rho, phi)
                # Convolution (unnecessary conjugate? map.data is real numbers.)
                c = (Z[self.notNan]*np.conjugate(self.data[self.notNan])).sum()
                cNorm = (Z[self.notNan]*np.conjugate(Z[self.notNan])).sum()
                c = c/cNorm
                tmp.append(c)
            A.append(tmp)
        if len(A) == 1:
            A = A[0]
            if len(A)==1:
                A = A[0]
                
        return A

    def rmZernikeCurvs(self, zModes='all',interpol=False):
        """
        Removes curvature from mirror map by fitting Zernike polynomials to the 
        mirror surface. Also stores the estimated radius of curvature
        and the Zernike polynomials and amplitudes.
        
        zModes   - 'defocus' uses only Zernike polynomial (n,m) = (2,0), which has a
                   parabolic shape.
                   'astigmatism' uses Zernike polynomials (n,m) = (2,-2) and (n,m) = (2,2).
                   'all' uses both defocus and the astigmatic modes.
        interpol - Boolean where 'true' means that the surface map is interpolated to get
                   more data points.

        Returns [Rc,A]
        Rc       - Estimated radius of curvature removed [m].
        A        - Amplitudes of Zernike polynomials removed [surfacemap.scaling].

        Based on the Simtools functions 'FT_remove_zernike_curvatures_from_map.m',
        'FT_zernike_map_convolution.m', etc. by Charlotte Bond.
        """
        
        tmp = deepcopy(self)
        ny,nx = self.data.shape
        # Interpolating for more precise convolution, in case size of map is small. 
        if interpol and (nx<1500 or ny<1500):
            # Number of extra steps inserted between two adjacent data points.
            N = math.ceil(1500.0/min(nx,ny))
            # New arrays of x-values and y-values
            x = np.arange(tmp.x[0],tmp.x[-1]+tmp.step_size[0]/(N+1),tmp.step_size[0]/N)
            y = np.arange(tmp.y[0],tmp.y[-1]+tmp.step_size[1]/(N+1),tmp.step_size[1]/N)

            """
            # --------------------------------------------------------------
            # Interpolating data
            tmp.interpolate(x,y)
            tmp.plot()
            # Interpolating for notNan.
            g = interp2d(self.x, self.y, self.notNan, kind='linear')
            tmp.notNan = np.round(g(x,y)) # round() or floor() here?! Can't decide...
            # Converting binary to boolean
            tmp.notNan = tmp.notNan==1
            # --------------------------------------------------------------
            """
            # Later when compatible with interpolation function, switch code
            # below for code above (but need to change 
            # --------------------------------------------------------------
            # Interpolation functions, for the data (f) and for notNan (g).
            f = interp2d(tmp.x, tmp.y, tmp.data, kind='linear')
            g = interp2d(tmp.x, tmp.y, tmp.notNan, kind='linear')
            # Interpolating
            tmp.data = f(x,y)

            tmp.notNan = np.round(g(x,y)) # round() or floor() here?! Can't decide...
            # Converting binary to boolean
            tmp.notNan = tmp.notNan==1

            # Setting new step size
            tmp.step_size = (x[1]-x[0],y[1]-y[0])
            # Setting new center
            fx = interp1d(x, np.arange(0,len(x)))
            fy = interp1d(y, np.arange(0,len(y)))
            tmp.center = (fx(0.0), fy(0.0))
            # --------------------------------------------------------------

        # Convolution between the surface map and Zernike polynomials
        # to get amplitudes of the polynomials. Only interested in n=2.
        A = tmp.zernikeConvol(2)[2]
        
        # Radius of mirror.
        R = self.find_radius(unit='meters')
        # Grid in polar coordinates
        rho,phi = self.createPolarGrid()
        rho = rho/R
        
        # Creates the choosen Zernike polynomials and removes them from the
        # mirror map.
        if zModes=='all' or zModes=='All':
            for k in range(0,3):
                m = (k-1)*2
                Z = A[k]*zernike(m, 2, rho, phi)
                self.data[self.notNan] = self.data[self.notNan]-Z[self.notNan]
                self.zernikeRemoved = (m, 2, A[k])
            # Estimating radius of curvature
            Rc = znm2Rc([a*self.scaling for a in A], R)
        elif zModes=='astigmatism' or zModes=='Astigmatism':
            for k in range(0,3,2):
                m = (k-1)*2
                Z = A[k]*zernike(m, 2, rho, phi)
                smap.data[self.notNan] = self.data[self.notNan]-Z[self.notNan]
                self.zernikeRemoved = (m, 2, A[k])
            Rc = znm2Rc([a*self.scaling for a in A[::2]], R)
        elif zModes=='defocus' or zModes=='Defocus':
            Z = A[1]*zernike(0, 2, rho, phi)
            self.data[self.notNan] = self.data[self.notNan]-Z[self.notNan]
            self.zernikeRemoved = (0, 2, A[1])
            Rc = znm2Rc(A[1]*self.scaling, R)
        
        self.Rc = Rc
        
        return self.Rc, self.zernikeRemoved


    def rmTilt(self, method='fitSurf', w=None, xbeta=None, ybeta=None, zOff=None):
        
        R = self.find_radius(unit='meters')
        
        if method=='zernike' or method=='Zernike':
            A1 = self.rmZernike(1, m = 'all')
            # Calculating equivalent tilts in radians
            xbeta = np.arctan(A1[1]*self.scaling/R)
            ybeta = np.arctan(A1[0]*self.scaling/R)
            self.betaRemoved = (xbeta,ybeta)
            return A1, xbeta, ybeta

        else:
            X,Y = np.meshgrid(self.x,self.y)
            r2 = X**2 + Y**2

            if w is not None:
                weight = 2/(math.pi*w**2) * np.exp(-2*r2[self.notNan]/w**2)

            def f(p):
                # This is used in simtools, why?
                #p[0] = p[0]*1.0e-9
                #p[1] = p[1]*1.0e-9
                Z = self.createSurface(0,X,Y,p[2],0,0,p[0],p[1])
                if w is None:
                    res = math.sqrt(((self.data[self.notNan]-Z[self.notNan])**2).sum())/self.notNan.sum()
                else:
                    # weight = 2/(math.pi*w**2) * np.exp(-2*r2[self.notNan]/w**2)
                    res = math.sqrt((weight*(self.data[self.notNan]-Z[self.notNan])**2).sum())/weight.sum()

                return res

            if xbeta is None:
                xbeta = 0
            if ybeta is None:
                ybeta = 0
            if zOff is None:
                zOff = 0

            params = [xbeta,ybeta,zOff]

            opts = {'xtol': 1.0e-8, 'ftol': 1.0e-8, 'maxiter': 2000, 'disp': False}
            out = minimize(f, params, method='Nelder-Mead', options=opts)

            xbeta = out['x'][0]
            ybeta = out['x'][1]
            zOff = out['x'][2]

            Z = self.createSurface(0,X,Y,zOff,0,0,xbeta,ybeta)
            self.data[self.notNan] = self.data[self.notNan] - Z[self.notNan]
            self.betaRemoved = (xbeta, ybeta)

            # Equivalent Zernike amplitude
            A1 = R*np.tan(np.array([ybeta,xbeta]))/self.scaling
            self.zernikeRemoved = (-1,1,A1[0])
            self.zernikeRemoved = (1,1,A1[1])
            
            return A1,xbeta,ybeta,zOff
        

    def rmSphericalSurf(self, Rc0, w=None, zOff=None, isCenter=[False,False]):
        """
        Fits spherical surface to the mirror map and removes it.

        Inputs: Rc0, w, zOff, isCenter
        Rc       - Initial guess of the Radius of curvature. [m]
        w        - Gaussian weighting parameter. The distance from the center where the
                   weigths have decreased by a factor of exp(-2) compared to the center.
                   Should preferrably be equal to the beam radius at the mirror. [m]
        zOff     - Initial guess of the z-offset. Generally unnecessary. [surfacemap.scaling]
        isCenter - 2D-list with booleans. isCenter[0] Determines if the center of the
                   sphere is to be fitted (True) or not (False, recommended). If the center is
                   fitted, then isCenter[1] determines if the weights (in case w!=None) are
                   centered around the fitted center (True) or centered around the center of
                   the data-grid (False, highly recommended).
                   
        if isCenter[0] == False
        Returns: Rc, zOff
        Rc       - Radius of curvature of the sphere removed from the mirror map. [m]
        zOff     - z-offset of the sphere removed. [surfacemap.scaling]
        isCenter[0] == True
        Returns Rc, zOff, center
        center   - [x0, y0] giving the coordinates of the center of the fitted sphere.
        
        Based on the file 'FT_remove_curvature_from_mirror_map.m' by Charlotte Bond.
        """
        
        # z-value at centre of the data-grid. Serves as initial guess for deviation
        # from z=0, if no other first guess is given.
        if zOff is None:
            zOff = self.data[round(self.center[1]), round(self.center[0])]

        # If fitting center of the sphere, four variables are fitted. Initial guess
        # of deviation from notNan-data-grid-center: (x0,y0) = (0,0).
        if isCenter[0]:
            p = [Rc0, zOff, 0.0, 0.0]
        # Otherwise two.
        else:
            p = [Rc0, zOff]

        # Grids with X,Y and r2 values. X and Y crosses zero in the center
        # of the xy-plane.
        X,Y = np.meshgrid(self.x, self.y)
        r2 = X**2 + Y**2

        # Cost-function to minimize.
        def costFunc(p):
            # If the center of the mirror is fitted, four variables are fitted.
            if isCenter[0]:
                Rc = p[0]
                zOff = p[1]
                x0 = p[2]
                y0 = p[3]
            # Otherwise 2 variables.
            else:
                Rc = p[0]
                zOff = p[1]
                x0 = 0
                y0 = 0

            Z = self.createSurface(Rc,X,Y,zOff,x0,y0)

            if w is None:
                # Mean squared difference between map and the created sphere.
                res = math.sqrt( ((self.data[self.notNan] -
                                   Z[self.notNan])**2).sum() )/self.notNan.sum()
            else:
                if isCenter[0] and isCenter[1]:
                    # Weights centered around fitting spehere center. May give weird
                    # results if the mirror deviates much from a sphere.
                    weight = (2/(math.pi*w**2))*np.exp(-2*( (X-x0)**2 + (Y-y0)**2 )/w**2)
                else:
                    # Weights centered around the center of the mirror xy-plane.
                    weight = (2/(math.pi*w**2))*np.exp(-2*r2/w**2)
                # Weighted mean squared difference between map and the created sphere.
                res = math.sqrt( ( weight[self.notNan]*
                                  ( (self.data[self.notNan] - Z[self.notNan])**2 )
                                  ).sum() ) \
                                  /weight[self.notNan].sum()
            return res

        # Using the simplex Nelder-Mead method. This is the same or very
        # similar to the method used in 'FT_remove_curvature_from_mirror_map.m',
        # but there are probably better methods to use.
        out = minimize(costFunc, p, method='Nelder-Mead',tol=1.0e-6)

        # Assigning values to the instance variables
        self.Rc = out['x'][0]
        if self.zOffset is None:
            self.zOffset = 0
        self.zOffset = self.zOffset + out['x'][1]

        # Equivalent Zernike (n=2,m=0) amplitude.
        R = self.find_radius(unit='meters')
        A20 = Rc2znm(self.Rc,R)/self.scaling
        self.zernikeRemoved = (0,2,A20)

        # If center was fitted, assign new values to instance variable center, and
        # subtract the fitted sphere from the mirror map.
        if isCenter[0]:
            x0 = out['x'][2]
            y0 = out['x'][3]
            # Converts the deviation into a new absolut center in data points.
            self.center = (self.center[0] + x0/self.step_size[0],
                           self.center[1] + y0/self.step_size[1])
            # Creating fitted sphere
            Z = self.createSurface(self.Rc, X, Y, self.zOffset, x0, y0)
            # Subtracting sphere from map
            self.data[self.notNan] = self.data[self.notNan]-Z[self.notNan]
            return self.Rc, self.zOffset, self.center, A20
        # Subtracting fitted sphere from mirror map.
        else:
            # Creating fitted sphere
            Z = self.createSurface(self.Rc,X,Y,self.zOffset)
            # Subtracting sphere from map
            self.data[self.notNan] = self.data[self.notNan]-Z[self.notNan]
            return self.Rc, self.zOffset, A20

    def remove_curvature(self, method='zernike', w=None, zOff=None,
                         isCenter=[False,False], zModes = 'all'):
        """
        Removes curvature from mirror map by fitting a sphere to the mirror map, or
        by convolving with second order Zernike polynomials and then extracting
        these polynomials with the obtained amplitudes. 

        Inputs: method, w, zOff, isCenter, zModes
        method   - 'zernike' (default) convolves second order Zernike polynomials with the
                   map, and then removes the polynomial with the obtained amplitude from the
                   surface map. Which of the three second order modes that are fitted is
                   determined by zMods (see below).
                 - 'sphere' fits a sphere to the mirror map. The fitting is Gaussian
                   weighted if w (see below) is specifed. 
        w        - Gaussian weighting parameter. The distance from the center where the
                   weigths have decreased by a factor of exp(-2) compared to the center.
                   Should preferrably be equal to the beam radius at the mirror. [m]
        zOff     - Initial guess of the z-offset. Only used together with method='sphere'.
                   Generally unnecessary. [surfacemap.scaling]
        isCenter - 2D-list with booleans. isCenter[0] Determines if the center of the
                   sphere is to be fitted (True) or not (False, recommended). If the center is
                   fitted, then isCenter[1] determines if the weights (in case w!=None) are
                   centered around the fitted center (True) or centered around the center of
                   the data-grid (False, highly recommended).
        zModes   - 'defocus' uses only Zernike polynomial (n,m) = (2,0), which has a
                   parabolic shape.
                   'astigmatism' uses Zernike polynomials (n,m) = (2,-2) and (n,m) = (2,2).
                   'all' uses both the defocus and the astigmatic modes.

        if method == 'zernike'
        Returns: Rc, znm
        Rc       - Approximate spherical radius of curvature removed.
        znm      - Dictionary specifying which Zernike polynomials that have been removed
                   as well as their amplitudes.
        if method == 'sphere' and isCenter[0] == False
        Returns: Rc, zOff
        Rc       - Radius of curvature of the sphere removed from the mirror map. [m]
        zOff     - z-offset of the sphere removed. [surfacemap.scaling]
        if method == 'sphere' and isCenter[0] == True
        Returns Rc, zOff, center
        center   - [x0, y0] giving the coordinates of the center of the fitted sphere. 
        """
        
        
        if method == 'zernike' or method == 'Zernike':
            Rc, znm = self.rmZernikeCurvs(zModes)
            return Rc, znm
        elif method == 'sphere' or method == 'Sphere':
            smap = deepcopy(self)
            # Using Zernike polynomial (m,n) = (0,2) to estimate radius of curvature.
            Rc0 = smap.rmZernikeCurvs('defocus')[0]
            if isCenter[0]:
                Rc, zOff, center, A20 = self.rmSphericalSurf(Rc0, w, zOff, isCenter)
                return Rc, zOff, self.center, A20
            else:
                Rc, zOff, A20 = self.rmSphericalSurf(Rc0, w, zOff, isCenter)
                return Rc, zOff, A20

            
    def recenter(self):
        """
        Setting center of mirror to be in the middle of the notNan field. 
        
        Based on 'FT_recenter_mirror_map.m' by Charlotte Bond.
        """
        # Row and column indices with non-NaN elements
        rIndx, cIndx = self.notNan.nonzero()
        # Finding centres
        x0 = float(cIndx.sum())/len(cIndx)
        y0 = float(rIndx.sum())/len(rIndx)
        self.center = tuple([x0,y0])
        return self.center
        # -------------------------------------------------
        
    def removePeriphery(self):
        """
        Finds the NaN elements closest to the map center, and sets all
        elements further out to NaN. Then cropping and recentering.
        
        Based on FT_remove_elements_outside_map.m by Charlotte Bond.
        """
        # Arrays of row and column indices where data is NaN.
        r,c = np.where(self.notNan == False)
        # Sets the map center to index [0,0]
        x = c-self.center[0]
        y = r-self.center[1]
        # Minimum radius squared measured in data points.
        r2 = (x**2+y**2).min()
        # Grid with the same dimensions as the map.
        X,Y = np.meshgrid(self.x, self.y)
        # Grid containing radial distances squared measured from the center.
        R2 = (X/self.step_size[0])**2 + (Y/self.step_size[1])**2
        # Matrix with True=='outside mirror surface'
        outs = R2>=r2
        # Setting notNan-matrix to false outside the mirror.
        self.notNan[outs] = False
        # Setting data matrix to 0 outside the mirror.
        self.data[outs] = 0.0

        # Removing unnecessary data points.
        # --------------------------------------
        # Radius inside data is kept. Don't know why R is set this way,
        # but trusting Dr. Bond.
        R = round(math.ceil(2.0*math.sqrt(r2)+6.0)/2.0)
        if 2*R<self.data.shape[0] and 2*R<self.data.shape[1]:
            x0 = round(self.center[0])
            y0 = round(self.center[1])

            self.data = self.data[y0-R:y0+R+1, x0-R:x0+R+1]
            self.notNan = self.notNan[y0-R:y0+R+1, x0-R:x0+R+1]
        
        self.recenter()
        
        
    
    def createSurface(self,Rc,X,Y,zOffset=0,x0=0,y0=0,xTilt=0,yTilt=0,isPlot=False):
        """
        Creating surface.
        
        Inputs: Rc, X, Y, zOffset=0, x0=0, y0=0, xTilt=0, yTilt=0, isPlot=False
        Rc      - Radius of curvature, and center of sphere on z-axis in case zOffset=0. [m]
        X, Y    - Matrices of x,y-values generated by 'X,Y = numpy.meshgrid(xVec,yVec)'. [m]
        zOffset - Surface center offset from 0. Positive value gives the surface center
                  positive z-coordinate. [self.scaling]
        x0,y0   - The center of the sphere in the xy-plane.
        x/yTilt - Tilt of x/y-axis [rad]. E.g. xTilt rotates around the y-axis.
        isPlot  - Plot or not [boolean]. Not recommended when used within optimization
                  algorithm.

        Returns: Z
        Z       - Matrix of surface height values. [self.scaling]
        
        Based on Simtools function 'FT_create_sphere_for_map.m' by Charlotte Bond.
        """
        
        # Adjusting for tilts and offset
        Z = zOffset + (X*np.tan(xTilt) + Y*np.tan(yTilt))/self.scaling
        # Adjusting for spherical shape.
        if Rc !=0 and Rc is not None:
            Z = Z + (Rc - np.sign(Rc)*np.sqrt(Rc**2 - (X-x0)**2 - (Y-y0)**2))/self.scaling
        
        if isPlot:
            import pylab
            pylab.clf()
            fig = pylab.figure()
            axes = pylab.pcolormesh(X, Y, Z) #, vmin=zmin, vmax=zmax)
            cbar = fig.colorbar(axes)
            cbar.set_clim(Z.min(), Z.max())
            pylab.title('Z_min = ' + str(Z.min()) + '   Z_max = ' + str(Z.max()))
            pylab.show()
        return Z
    
    def createPolarGrid(self):
        """
        Creating polar grid from the rectangular x and y coordinates in
        surfacemap. 

        Returns: rho, phi
        rho  - matrix with radial distances from centre of mirror map.
        phi  - matrix with polar angles.
        """
        X,Y = np.meshgrid(self.x,self.y)
        phi = np.arctan2(Y,X)
        rho = np.sqrt(X**2 + Y**2)
        return rho, phi

    def preparePhaseMap(self, xyOffset=None, w=None, verbose=False):
        """
        Based on Simtools function 'FT_prepare_phase_map_for_finesse.m' by Charlotte Bond.
        """
        if verbose:
            print('--------------------------------------------------')
            print('Preparing phase map for Finesse...')
            print('--------------------------------------------------')
        w_max = self.find_radius(method='min', unit='meters')
        if w is None:
            if verbose:
                print(' No weighting used')
        elif w >= w_max:
            if verbose:
                print(' Weighting radius ({:.3f} cm) larger than mirror radius ({:.3f} cm).'.format(w*100, w_max*100))
                print(' Thus, no weighting used.')
            w = None
        elif verbose:
            print(' Gaussian weights used with radius: {:.2f} cm'.format(w*100.0))
        if verbose:
            print(' (rms, avg) = ({:.3e}, {:.3e}) nm'.format(self.rms(w),self.avg(w)))
            print('--------------------------------------------------')
        
            print(' Centering...')
        self.recenter()
        if verbose:
            print('  New center (x0, y0) = ({:.2f}, {:.2f})'.
                  format(self.center[0],self.center[1]))
        if verbose:
            print(' Cropping map...')
        before = self.size
        self.removePeriphery()
        if verbose:
            print('  Size (rows, cols): ({:d}, {:d}) ---> ({:d}, {:d})'.
                  format(before[1], before[0], self.size[1], self.size[0]))
            print('  New center (x0, y0) = ({:.2f}, {:.2f})'.
                  format(self.center[0],self.center[1]))
            print('  (rms, avg) = ({:.3e}, {:.3e}) nm'.
                  format(self.rms(w),self.avg(w)))

        # Radius of mirror in xy-plane.
        R = self.find_radius(unit='meters')
        
        if verbose:
            print(' Removing curvatures...')
        # --------------------------------------------------------
        if w is None:
            Rc, znm = self.remove_curvature(method='zernike', zModes = 'defocus')
            if verbose:
                print('  Removed Z20 with amplitude A20 = {:.2f} nm'.format(znm['02'][2]))
                print('  Equivalent Rc = {:.2f} m'.format(Rc))
        else:
            Rc, zOff, A20 = self.remove_curvature(method='sphere', w=w)
            # Adding A20 to zOff (calling it A0 now) because: Z(n=2,m=0) = A20*(r**2 - 1)
            A0 = zOff+A20
            if verbose:
                print('  Removed Rc = {0:.2f} m'.format(Rc) )
                print('  Equivalent Z(n=2,m=0) amplitude A20 = {0:.2f} nm'.format(A20))

        if verbose:
            print('  (rms, avg) = ({:.3e}, {:.3e}) nm'.format(self.rms(w),self.avg(w)))
            
            print(' Removing offset...')
        # --------------------------------------------------------

        zOff = self.removeOffset(w)
        if w is None:
            if verbose:
                print('  Removed Z00 with amplitude A00 = {:.2f} nm'.format(zOff) )
        else:
            A0 = A0 + zOff
            if verbose:
                print('  Removed z-offset (A00) = {0:.3f} nm'.format(zOff))
        if verbose:
            print('  (rms, avg) = ({:.3e}, {:.3e}) nm'.format(self.rms(w),self.avg(w)))
        
            print(' Removing tilts...')
        # --------------------------------------------------------
        if w is None:
            A1,xbeta,ybeta = self.rmTilt(method='zernike')
            if verbose:
                for k in range(len(A1)):
                    print('  Removed Z1{:d} with amplitude A1{:d} = {:.2f} nm'.
                          format(2*k-1,2*k-1,A1[k]) )
                print('  Equivalent tilts in radians: ')
                print('   xbeta = {:.2e} rad'.format(xbeta))
                print('   ybeta = {:.2e} rad'.format(ybeta))
        else:
            A1,xbeta,ybeta,zOff = self.rmTilt(method='fitSurf', w=w)
            A0 = A0 + zOff
            if verbose:
                print('  Tilted surface removed:')
                print('   xbeta    = {:.2e} rad'.format(xbeta))
                print('   ybeta    = {:.2e} rad'.format(ybeta))
                print('   z-offset = {:.2e} nm'.format(zOff))
                print('  Equivalent Zernike amplitudes:')
                print('   A(1,-1) = {:.2f} nm'.format(A1[0]))
                print('   A(1, 1) = {:.2f} nm'.format(A1[1]))
        if verbose:
            print('  (rms, avg) = ({:.3e}, {:.3e}) nm'.format(self.rms(w),self.avg(w)))

        if w is not None:
            # Adding the total offset removed to zernikeRemoved (just for the record)
            self.zernikeRemoved = (0,0,A0)
            if verbose:
                print(' Equivalent Z00 amplitude from accumulated z-offsets, A00 = {:.2f}'.format(A0))
            
        if xyOffset is not None:
            if verbose:
                print(' Offsetting mirror center in the xy-plane...')
            # --------------------------------------------------------
            self.center = (self.center[0] + xyOffset[0]/self.step_size[0],
                           self.center[1] + xyOffset[1]/self.step_size[1])
            self.xyOffset = xyOffset
            if verbose:
                print('  New mirror center (x0, y0) = ({:.2f}, {:.2f})'.format(self.center[0],self.center[1]))
        else:
            self.xyOffset = (0,0)
            
        if verbose:
            print(' Writing phase map to file...')
        # --------------------------------------------------------
        filename = self.name + '_finesse.txt'
        self.write_map(filename)
        if verbose:
            print('  Phase map written to file {:s}'.format(filename))
        
        
        print(' Writing result information to file...')
        # --------------------------------------------------------
        filename = self.name + '_finesse_info.txt'
        self.writeResults(filename)
        print('  Result written to file {:s}'.format(filename))
        

        if verbose:
            print('--------------------------------------------------')
            
        print('--------------------------------------------------')
        print('Phase map prepared! :D' )
        print('Name: {:s}'.format(self.name))
        print('Type: {:s}'.format(self.type))
        print('--------------------------------------------------')
            
        print('Radius:    R     = {:.2f} cm'.format(self.find_radius(unit='meters')*100))
        print('Offset:    A00   = {:.2f} nm'.format(self.zernikeRemoved['00'][2]))
        print('Tilt:      A1-1  = {:.2f} nm'.format(self.zernikeRemoved['-11'][2]))
        print('           A11   = {:.2f} nm,  or'.format(self.zernikeRemoved['11'][2]))
        print('           xbeta = {:.2e} rad'.format(self.betaRemoved[0]))
        print('           ybeta = {:.2e} rad'.format(self.betaRemoved[1]))
        print('Curvature: A20   = {:.2f} nm,  or'.format(self.zernikeRemoved['02'][2]))
        print('           Rc    = {:.2f} m'.format(self.Rc))
        print('xy-offset: x0    = {:.2f} mm'.format(self.xyOffset[0]*1000))
        print('           y0    = {:.2f} mm'.format(self.xyOffset[1]*1000))
        print('--------------------------------------------------')
            
        
        
        print()

        # Todo:
        # Add "create aperture map"
        
    
    def writeResults(self, filename):
        """
        Writing results to file. Not yet finished.
        """
        import time
        with open(filename,'w') as mapfile:
            mapfile.write('---------------------------------------\n')
            mapfile.write('Map:  {:s}\n'.format(self.name))
            mapfile.write('Date:  {:s}\n'.format(time.strftime("%d/%m/%Y  %H:%M:%S")))
            mapfile.write('---------------------------------------\n')
            mapfile.write('Diameter:  {:.2f} cm\n'.format(self.find_radius(unit='meters')*200.0))
            mapfile.write('Offset:    A00   = {:.2f} nm\n'.format(self.zernikeRemoved['00'][2]))
            mapfile.write('Tilt:      A1-1  = {:.2f} nm\n'.format(self.zernikeRemoved['-11'][2]))
            mapfile.write('           A11   = {:.2f} nm,  or\n'.format(self.zernikeRemoved['11'][2]))
            mapfile.write('           xbeta = {:.2e} rad\n'.format(self.betaRemoved[0]))
            mapfile.write('           ybeta = {:.2e} rad\n'.format(self.betaRemoved[1]))
            mapfile.write('Curvature: A20   = {:.2f} nm,  or\n'.format(self.zernikeRemoved['02'][2]))
            mapfile.write('           Rc    = {:.2f} m\n'.format(self.Rc))
            mapfile.write('xy-offset: x0    = {:.2f} mm\n'.format(self.xyOffset[0]*1000))
            mapfile.write('           y0    = {:.2f} mm\n'.format(self.xyOffset[1]*1000))
            
            # mapfile.write('Offset (Z00):  {:.2f}'.format(self.zernikeRemoved['00'][2]))
          
        
    def removeOffset(self, r=None):
        """
        Based on Simtools function 'FT_remove_offset_from_mirror_map.m' by Charlotte Bond.
        """
        if r is None:
            offset = self.rmZernike(0,0)
        else:
            rho = self.createPolarGrid()[0]
            inside = rho<r
            inside_notNan = np.zeros(inside.shape,dtype=bool)
            inside_notNan[inside] = self.notNan[inside]

            offset = self.data[inside_notNan].sum()/inside_notNan.sum()
            self.data[self.notNan] = self.data[self.notNan]-offset
        
        return offset

    def rmZernike(self, n, m='all'):
        """
        Removes Zernike polynomials from mirror map.

        Input: n, m
        n  - Radial degree of Zernike polynomial.
        m  - Azimuthal degree of Zernike polynomial. If set to 'all', all polynomials
             of radial degree n is removed. Can also be an integer, or a list of
             specifying the Azimuthal degrees that will be removed. Must be consistent
             with the chosen radial degree.

        Returns: A
        A  - List of amplitudes of the removed Zernike polynomials.
        """
        if m=='all':
            m = range(-n,n+1,2)
            
        R = self.find_radius(unit='meters')
        A = self.zernikeConvol(n,m)
        rho, phi= self.createPolarGrid()
        rho = rho/R

        if isinstance(m,list):
            for k in range(len(m)):
                Z = zernike(m[k], n, rho, phi)
                self.data[self.notNan] = self.data[self.notNan]-A[k]*Z[self.notNan]
                self.zernikeRemoved = (m[k], n, A[k])
        else:
            Z = zernike(m, n, rho, phi)
            self.data[self.notNan] = self.data[self.notNan]-A*Z[self.notNan]
            self.zernikeRemoved = (m, n, A)
        return A
        
class mergedmap:
    """
    A merged map combines multiple surfaces map to form one. Such a map can be used
    for computations of coupling coefficients but it cannot be written to a file to 
    be used with Finesse. For this you must output each map separately.
    
    """
    
    def __init__(self, name, size, center, step_size, scaling):
        
        self.name = name
        self.center = center
        self.step_size = step_size
        self.scaling = scaling
        self.__interp = None
        self._rom_weights = None
        self.__maps = []
        self.weighting = None
        
    def addMap(self, m):
        self.__maps.append(m)
    
    @property
    def center(self):
        return self.__center
    
    @center.setter
    def center(self, value):
        self.__center = value
        self.__interp = None
    
    @property
    def type(self):
        hasR = False
        hasT = False
        
        _type = ""
        
        for m in self.__maps:
            if "reflection" in m.type: hasR = True
            
            if "transmission" in m.type: hasT = True
            
            if "both" in m.type:
                hasR = True
                hasT = True
        
        if hasR and not hasT: _type += "reflection "
        elif hasR and not hasT: _type += "transmission "
        elif hasR and hasT: _type += "both "
        
        return _type
        
    @property
    def step_size(self):
        return self.__step_size
    
    @step_size.setter
    def step_size(self, value):
        self.__step_size = value
        self.__interp = None

    @property
    def scaling(self):
        return self.__scaling
    
    @scaling.setter
    def scaling(self, value):
        self.__scaling = value
        self.__interp = None
    
    @property
    def x(self):
        return self.step_size[0] * (np.array(range(0, self.size[0])) - self.center[0])
        
    @property
    def y(self):
        return self.step_size[1] * (np.array(range(0, self.size[1])) - self.center[1])

    @property
    def size(self):
        return self.__maps[0].data.shape
            
    @property
    def offset(self):
        return np.array(self.step_size)*(np.array(self.center) - (np.array(self.size)-1)/2.0)
    
    @property
    def ROMWeights(self):
        return self._rom_weights
    
    def z_xy(self, wavelength=1064e-9, direction="reflection_front", nr1=1.0, nr2=1.0):
        
        z_xy = np.ones(self.size, dtype=np.complex128)
        
        for m in self.__maps:
            z_xy *= m.z_xy(wavelength=wavelength, direction=direction, nr1=nr1, nr2=nr2)
            
        if self.weighting is None:
            return z_xy
        else:
            return z_xy * self.weighting
        
    def generateROMWeights(self, EIxFilename, EIyFilename=None, verbose=False, interpolate=False, newtonCotesOrder=8, nr1=1, nr2=1):
        if interpolate == True:
            # Use EI nodes to interpolate if we
            with open(EIxFilename, 'rb') as f:
                EIx = pickle.load(f)

            if EIyFilename is None:
                EIy = EIx
            else:
                with open(EIyFilename, 'rb') as f:
                    EIy = pickle.load(f)

            x = EIx.x
            x.sort()
            nx = np.unique(np.hstack((x, -x[::-1])))
        
            y = EIy.x
            y.sort()
            ny = np.unique(np.hstack((y, -y[::-1])))
            
            self.interpolate(nx, ny)
        
        w_refl_front, w_refl_back, w_tran_front, w_tran_back = (None, None, None, None)
        
        if "reflection" in self.type or "both" in self.type:
            w_refl_front = makeWeightsNew(self, EIxFilename, EIyFilename,
                                      verbose=verbose, newtonCotesOrderMapWeight=newtonCotesOrder,
                                      direction="reflection_front")
            
            w_refl_front.nr1 = nr1
            w_refl_front.nr2 = nr2
            
            w_refl_back = makeWeightsNew(self, EIxFilename, EIyFilename,
                                      verbose=verbose, newtonCotesOrderMapWeight=newtonCotesOrder,
                                      direction="reflection_back")
            
            w_refl_back.nr1 = nr1
            w_refl_back.nr2 = nr2

        if "transmission" in self.type or "both" in self.type:                                      
            w_tran_front = makeWeightsNew(self, EIxFilename, EIyFilename,
                                      verbose=verbose, newtonCotesOrderMapWeight=newtonCotesOrder,
                                      direction="transmission_front")

            w_refl_front.nr1 = nr1
            w_refl_front.nr2 = nr2
                                            
            w_tran_back  = makeWeightsNew(self, EIxFilename, EIyFilename,
                                      verbose=verbose, newtonCotesOrderMapWeight=newtonCotesOrder,
                                      direction="transmission_back")
            
            w_refl_back.nr1 = nr1
            w_refl_back.nr2 = nr2
            
        self._rom_weights = MirrorROQWeights(w_refl_front, w_refl_back, w_tran_front, w_tran_back)
        
        return self._rom_weights

    def interpolate(self, nx, ny, **kwargs):
        """
        Interpolates all the maps that are used to fc
        
        Uses scipy.interpolate.interp2d and any keywords arguments are
        passed on to it, thus settings like interpolation type and
        fill values can be set.
        
        The range of nx and ny must contain the value zero so that the
        center point of the map can be set.
        """

        for m in self.__maps:
            m.interpolate(nx, ny)

    def plot(self, mode="absorption", show=True, clabel=None, xlim=None, ylim=None, wavelength=1064e-9):
        
        import pylab
        
        if xlim is not None:
            _x = np.logical_and(self.x<=max(xlim)/100.0, self.x>=min(xlim)/100.0)
            xmin = np.min(np.where(_x == True))
            xmax = np.max(np.where(_x == True))
        else:
            xmin = 0
            xmax = len(self.x)-1
            xlim = [self.x.min()*100, self.x.max()*100]
    
        if ylim is not None:
            _y = np.logical_and(self.y<=max(ylim)/100.0, self.y>=min(ylim)/100.0)
            ymin = np.min(np.where(_y == True))
            ymax = np.max(np.where(_y == True))
        else:
            ymin = 0
            ymax = len(self.y)-1
            ylim = [self.y.min()*100, self.y.max()*100]

        if mode == "absorption":
            # plots how much of field is absorbed
            data = 1-np.abs(self.z_xy())
        elif mode == "meter":
            # plot the phase in terms of meters of displacement
            k = 2*np.pi/wavelength
            data = np.angle(self.z_xy()) / (2*k)
            
        zmin = data[xmin:xmax,ymin:ymax].min()
        zmax = data[xmin:xmax,ymin:ymax].max()

        # 100 factor for scaling to cm
        xrange = 100*self.x
        yrange = 100*self.y

        fig = pylab.figure()
        axes = pylab.pcolormesh(xrange, yrange, data, vmin=zmin, vmax=zmax)
        pylab.xlabel('x [cm]')
        pylab.ylabel('y [cm]')

        if xlim is not None: pylab.xlim(xlim)
        if ylim is not None: pylab.ylim(ylim)

        pylab.title('Merged map {0}, mode {1}'.format(self.name, mode))

        cbar = fig.colorbar(axes)
        cbar.set_clim(zmin, zmax)
        
        if clabel is not None:
            cbar.set_label(clabel)
    
        if show:
            pylab.show()
        
        return fig

class aperturemap(surfacemap):
    
    def __init__(self, name, size, step_size, R):
        surfacemap.__init__(self, name, "absorption both", size, (np.array(size)+1)/2.0, step_size, 1)
        self.R = R
        
    @property
    def R(self):
        return self.__R
    
    @R.setter
    def R(self, value):
        self.__R = value
    
        xx, yy = np.meshgrid(self.x, self.y)
        
        radius = np.sqrt(xx**2 + yy**2)
        
        self.data = np.zeros(self.size)
        self.data[radius > self.R] = 1.0
        
        
class curvedmap(surfacemap):
    
    def __init__(self, name, size, step_size, Rc):
        surfacemap.__init__(self, name, "phase reflection", size, (np.array(size)+1)/2.0, step_size, 1e-6)
        self.Rc = Rc
        
    @property
    def Rc(self):
        return self.__Rc
    
    @Rc.setter
    def Rc(self, value):
        self.__Rc = value
    
        xx, yy = np.meshgrid(self.x, self.y)
        
        Rsq = xx**2 + yy**2
        self.data = (self.Rc - math.copysign(1.0, self.Rc) * np.sqrt(self.Rc**2 - Rsq))/ self.scaling

class tiltmap(surfacemap):
    """
    To create a tiltmap, plot it and write it to a file to use with Finesse:
        
        tilts = (1e-6, 1e-8) # tilt in (x, y) radians\
        dx = 1e-4
        L = 0.2
        N = L/dx
        
        tmap = tiltmap("tilt", (N, N), (dx,dx), tilts)
        tmap.plot()
        tmap.write_map("mytilt.map")
    """
    
    def __init__(self, name, size, step_size, tilt):
        surfacemap.__init__(self, name, "phase reflection", size, (np.array(size)+1)/2.0, step_size, 1e-9)
        self.tilt = tilt
        
    @property
    def tilt(self):
        return self.__tilt
    
    @tilt.setter
    def tilt(self, value):
        self.__tilt = value
        
        xx, yy = np.meshgrid(self.x, self.y)
        
        self.data = (yy * self.tilt[1] + xx * self.tilt[0])/self.scaling
        

class zernikemap(surfacemap):
	def __init__(self, name, size, step_size, radius, scaling=1e-9):
		surfacemap.__init__(self, name, "phase reflection", size, (np.array(size)+1)/2.0, step_size, scaling)
		self.__zernikes = {}
		self.radius = radius
		
	@property
	def radius(self): return self.__radius

	@radius.setter
	def radius(self, value, update=True):
		self.__radius = float(value)
		if update: self.update_data()

	def setZernike(self, m, n, amplitude, update=True):
		self.__zernikes["%i%i" % (m, n)] = (m,n,amplitude)
		if update: self.update_data()

	def update_data(self):
		X,Y = np.meshgrid(self.x, self.y)
		R = np.sqrt(X**2 + Y**2)
		PHI = np.arctan2(Y, X)

		data = np.zeros(np.shape(R))

		for i in self.__zernikes.items():
			data += i[1][2] * zernike(i[1][0], i[1][1], R/self.radius, PHI)

		self.data = data
	
			

def read_map(filename, mapFormat='finesse', scaling=1.0e-9):
    """
    Reads surface map files and return a surfacemap-object xy-centered at the
    center of the mirror surface.
    filename  - name of surface map file.
    mapFormat - 'finesse', 'ligo', 'zygo'. Currently only for ascii formats.
    scaling   - scaling of surface height of the mirror map [m].
    """
    
    # Function converting string number into float.
    g = lambda x: float(x)

    # Reads finesse mirror maps.
    if mapFormat == 'finesse':
        
        with open(filename, 'r') as f:
        
            f.readline()
            name = f.readline().split(':')[1].strip()
            maptype = f.readline().split(':')[1].strip()
            size = tuple(map(g, f.readline().split(':')[1].strip().split()))
            center = tuple(map(g, f.readline().split(':')[1].strip().split()))
            step = tuple(map(g, f.readline().split(':')[1].strip().split()))
            scaling = float(f.readline().split(':')[1].strip())
        
        data = np.loadtxt(filename, dtype=np.float64,ndmin=2,comments='%')    

        return surfacemap(name, maptype, size, center, step, scaling, data)
        
    # Converts raw zygo and ligo mirror maps to the finesse
    # format. Based on the matlab scripts 'FT_read_zygo_map.m' and
    # 'FT_read_ligo_map.m'.
    elif mapFormat == 'ligo' or mapFormat == 'zygo':
        if mapFormat == 'ligo':
            isLigo = True
            # Remove '_asc.dat' for output name
            name = filename.split('_')
            name = '_'.join(name[:-1])
        else:
            isLigo = False
            tmp = filename.split('.')
            fileFormat = tmp[-1].strip()
            name = '.'.join(tmp[:-1])
            if fileFormat == 'asc':
                isAscii = True
            else:
                isAscii = False
                
        # Unknowns (why are these values hard coded here?) Moving
        # scaling to input. Fix the others later too.
        # ------------------------------------------------------
        # Standard maps have type 'phase' (they store surface
        # heights)
        maptype = 0
        # Both (reflected and transmitted) light fields are
        # affected
        field = 0
        # Measurements in nanometers
        # scaling = 1.0e-9
        # ------------------------------------------------------

        # Reading header of LIGO-map (Zygo file? Says Zygo in
        # header...)
        # ------------------------------------------------------
        with open(filename, 'r') as f:
            # Skip first two lines
            for k in range(2):
                f.readline()
            # If zygo-file, and ascii format, there is intensity
            # data. Though, the Ligo files I have seen are also
            # zygo-files, so maybe we should extract this data
            # from these too?
            line = f.readline()
            if not isLigo and isAscii:
                iCols = float(line.split()[2])
                iRows = float(line.split()[3])
                
            line = f.readline().split()
            # Unknown usage.
            # ----------------------------------------------
            if isLigo:
                y0 = float(line[0])
                x0 = float(line[1])
                rows = float(line[2])
                cols = float(line[3])
            else:
                y0 = float(line[1])
                x0 = float(line[0])
                rows = float(line[3])
                cols = float(line[2])
            # ----------------------------------------------

            # Skipping three lines
            for k in range(3):
                f.readline()
            line = f.readline().split()

            # Unknown (Scaling factors)
            # ----------------------------------------------
            # Interfeometric scaling factor (?)
            S = float(line[1])
            # wavelength (of what?)
            lam = float(line[2])
            # Obliquity factor (?)
            O = float(line[4])
            # ----------------------------------------------
            # Physical step size in metres
            if line[6] != 0:
                xstep = float(line[6])
                ystep = float(line[6])
            else:
                xstep = 1.0
                ystep = 1.0
                
            # Skipping two lines
            for k in range(2):
                f.readline()
            line = f.readline().split()

            # Unknown
            # Resolution of phase data points, 1 or 0.
            phaseRes = float(line[0])
            if phaseRes == 0:
                R = 4096
            elif phaseRes == 1:
                R = 32768
            else:
                print('Error, invalid phaseRes')

            if not isLigo and not isAscii:
                # zygo .xyz files give phase data in microns.
                hScale = 1.0e-6
            else:
                # zygo .asc and ligo-files give phase data in
                # internal units. To convert to m use hScale
                # factor.
                hScale = S*O*lam/R
                
            if not isLigo and not isAscii:
                print('Not implemented yet, need a .xyz-file ' +
                      'to do this.')
                return 0
                
            # Skipping four lines
            for k in range(4):
                f.readline()
            if not isLigo and isAscii:
                # Reading intensity data
                iData = np.array([])
                line = f.readline().split()
                while line[0] != '#':
                    iData = np.append(iData, map(g,line))
                    line = f.readline().split()
                # Reshaping intensity data
                iData = iData.reshape(iRows, iCols).transpose()
                iData = np.rot90(iData)
            else:
                # Skipping lines until '#' is found.
                while f.readline()[0] != '#':
                    pass
                
            # Reading phase data
            # ----------------------------------------------
            # Array with the data
            data = np.array([])
            # Reading data until next '#' is reached.
            line = f.readline().split()
            while line[0] != '#':
                data = np.append(data, map(g,line))
                line = f.readline().split()
            # ----------------------------------------------

        
        if isLigo:
            # Setting all the points outside of the mirror
            # surface to NaN. These are given a large number
            # in the file. 
            data[data == data[0]] = np.nan
            
            # Reshaping into rows and columns
            data = data.reshape(cols,rows).transpose()
            # Pretty sure that the lines below can be done
            # more efficient, but it's quick as it is.
            # ----------------------------------------------
            # Flipping right and left
            data = np.fliplr(data)
            # Rotating 90 degrees clockwise 
            data = np.rot90(data,-1)
            # Flipping right and left
            data = np.fliplr(data)
            # ----------------------------------------------
        else:
            if isAscii:
                # Setting all the points outside of the mirror
                # surface to NaN. These are given a large number
                # in the file. 
                data[data >= 2147483640] = np.nan
            # Reshaping into rows and columns.
            data = data.reshape(rows,cols).transpose()
            # Rotating to make (0,0) be in bottom left
            # corner. 
            data = np.rot90(data)
            
        # Scaling to nanometer (change this to a user
        # defined value?) Still don't know where
        # 'hScale' really comes from.
        data = (hScale/scaling)*data
        size = data.shape

        if maptype == 0:
            mType = 'phase'
        else:
            mType = 'Unknown'
        if field == 0:
            fType = 'both'
        else:
            fType = 'unknown'

        maptype = ' '.join([mType, fType])

        # Wrong! fix by creating recenter method.
        center = tuple([x0,y0])
        step = tuple([xstep,ystep])

        # Simple re-centering of mirror, translated from
        # 'FT_recenter_mirror_map.m'
        # -------------------------------------------------
        # Matrix with ones where data element is not NaN.
        isNan = np.isnan(data)
        notNan = isNan==False
        # Row and column indices with non-NaN elements
        rIndx, cIndx = notNan.nonzero()
        # Finding centres
        x0 = float(cIndx.sum())/len(cIndx)
        y0 = float(rIndx.sum())/len(rIndx)
        center = tuple([x0,y0])
        # -------------------------------------------------
        
        # Changing NaN to zeros. Just to be able to plot the
        # map with surfacemap.plot().
        data[isNan] = 0 
        
        return surfacemap(name, maptype, (0,0), center, step, scaling, data, notNan)

    if mapFormat == 'metroPro':
        # Remove '.dat' for output name
        name = filename.split('.')[0]
        # Reading file
        data,hData,x1,y1 = readMetroProData(filename)
        # xy-stepsize in meters
        stepSize = (hData['cameraRes'], hData['cameraRes'])
        # Temporary center
        center = ((len(data[0,:])-1)/2.0,(len(data[:,0])-1)/2.0)
        # Keeping track of nan's
        isNan = np.isnan(data)
        notNan = isNan==False
        data[isNan] = 0
        # Change this:
        mType = 'phase'
        fType = 'both'
        maptype = ' '.join([mType, fType])
        # Unnecessary size, remove from class later
        size = data.shape[::-1]
        smap = surfacemap(name, maptype, size, center, stepSize, 1.0, data, notNan)
        smap.rescaleData(scaling)
        smap.recenter()
        
        return smap
'''
    
def readMetroProData(filename):
    '''
    Reading the metroPro binary data files.

    Translated from 'LoadMetroProData.m' by Hiro Yamamoto.
    '''
    f = BinaryReader(filename)
    # Read header
    hData = readHeaderMP(f)
    if hData['format'] < 0:
        print('Error: Format unknown to readMetroProData()\nfilename: {:s}'.format(filename))
        return 0
    # Read phase map data
    # Skipping header and intensity data
    f.seek(hData['size']+hData['intNBytes'])
    # Reading data
    dat = f.read('int32',size=hData['Nx']*hData['Ny'])
    # Marking unmeasured data as NaN
    dat[dat >= hData['invalid']] = np.nan
    # Scale data to meters
    dat = dat*hData['convFactor']
    # Reshaping into Nx * Ny matrix
    dat = dat.reshape(hData['Ny'], hData['Nx'])
    # Flipping up/down, i.e., change direction of y-axis.
    dat = dat[::-1,:]
    # Auxiliary data to return
    dxy = hData['cameraRes']
    if dxy == 0.0:
        print('WARNING: Unable to read pixel scale, set to unit.')
        dxy = 1.  # unknowm, set to 1 (pixel)
    
    x1 = dxy * np.arange( -(len(dat[0,:])-1)/2, (len(dat[0,:])-1)/2 + 1)
    y1 = dxy * np.arange( -(len(dat[:,0])-1)/2, (len(dat[:,1])-1)/2 + 1)

    return dat, hData, x1, y1


def readHeaderMP(f):
    '''
    Reads header of the metroPro format files.
    
    Translated from the readerHeader() function within the
    'LoadMetroProData.m' function written by Hiro Yamamoto.
    '''
    hData = {}
    hData['magicNum'] = f.read('uint32')
    hData['format'] = f.read('int16')
    hData['size'] = f.read('int32')
    # Check if the magic string and format are known ones.
    # from MxReference 0550_C:
    #There are only three recognized combinations of these values as shown in the following table:
    #magic_number (hex) header_format header_size (bytes)
    #0x881B036F 1 834
    #0x881B0370 2 834
    #0x881B0371 3 4096
    #Reader software that is rigorous should read all three values and validate them per the above
    #table.
    if not (hData['format'] >= 1 and hData['format']<=3 and
            hData['magicNum']-hData['format'] == int('881B036E',16)):
        hData['format'] = -1
    # Read necessary data
    hData['invalid'] = int('7FFFFFF8',16)
    f.seek(60)
    # Intensity data, which we will skip over.
    hData['intNBytes'] = f.read('int32')
    # Top-left coordinates, which are useless.
    hData['X0'] = f.read('int16')
    hData['Y0'] = f.read('int16')
    # Number of data points alng x and y
    hData['Nx'] = f.read('int16')
    hData['Ny'] = f.read('int16')
    # Total data, 4*Nx*Ny
    hData['phaNBytes'] = f.read('int32')
    f.seek(218)
    # Scale factor determined by phase resolution tag
    phaseResTag = f.read('int16')
    if phaseResTag == 0:
        phaseResVal = 4096
    elif phaseResTag == 1:
        phaseResVal = 32768
    elif phaseResTag == 2:
        phaseResVal = 131072
    else:
        phaseResVal = 0
    f.seek(164)
    intf_scale_factor = f.read('float')
    hData['waveLength'] = f.read('float')
    f.seek(176)
    obliquity_factor = f.read('float')
    # Eq. in p12-6 in MetroPro Reference Guide
    hData['convFactor'] = intf_scale_factor*obliquity_factor*\
                          hData['waveLength']/phaseResVal
    # Bin size of each measurement
    f.seek(184)
    hData['cameraRes'] = f.read('float')
    # camera resolution in meters/pixel (unknown if zero). Multiply this number by the appropriate factor to obtain the resolution needed. The horizontal and vertical pixel resolution can be calculated as follows:
    # CameraRes = 1.81512e-005, to get microns (10-6 meters): (.000018512)*(1000000) = 18.512 microns. The horizontal and vertical pixel resolutions are the same. 
    
    return hData


import struct

class BinaryReaderEOFException(Exception):
    def __init__(self):
        pass
    def __str__(self):
        return 'Not enough bytes in file to satisfy read request'

class BinaryReader:
    # Map well-known type names into struct format characters.
    typeNames = {
        'int8'   :'b',
        'uint8'  :'B',
        'int16'  :'h',
        'uint16' :'H',
        'int32'  :'i',
        'uint32' :'I',
        'int64'  :'q',
        'uint64' :'Q',
        'float'  :'f',
        'double' :'d',
        'char'   :'s'}

    def __init__(self, fileName):
        self.file = open(fileName, 'rb')
        
    def read(self, typeName, size=None):
        typeFormat = BinaryReader.typeNames[typeName.lower()]
        typeFormat = '>'+typeFormat
        typeSize = struct.calcsize(typeFormat)
        if size is None:
            value = self.file.read(typeSize)
            if typeSize != len(value):
                raise BinaryReaderEOFException
            unpacked = struct.unpack(typeFormat, value)[0]
        else:
            value = self.file.read(size*typeSize)
            if size*typeSize != len(value):
                raise BinaryReaderEOFException
            unpacked = np.zeros(size)
            for k in range(size):
                i = k*typeSize
                unpacked[k] = struct.unpack(typeFormat,value[i:i+typeSize])[0]
        return unpacked

    def seek(self, offset, refPos=0):
        '''
        offset in bytes and refPos gives reference position, where 0
        means origin of the file, 1 uses current position and 2 uses
        the end of the file.
        '''
        self.file.seek(offset, refPos)
    
    
    def __del__(self):
        self.file.close()

        
# TODO: Add options for reading virgo maps, and .xyz zygo
    # maps (need .xys file for this). Binary ligo-maps?
    # The intensity data is not used to anything here. Remove
    
    
    # or add to pykat?
    
    
    
    
# TODO: Recreate functions from Simtools:, List taken from: ligo_maps/FT_convert_ligo_map_for_finesse.m
# map=FT_recenter_mirror_map(map);
# [map2,A2,Rc_out]=FT_remove_zernike_curvatures_from_map(map,Rc_in);
# [map2,Rc_out]=FT_remove_curvature_from_mirror_map(map,Rc_in,w, display_style);
# [map2,offset]=FT_remove_offset_from_mirror_map(map2,1e-2);
# [map3,x_tilt,y_tilt,offset2]=FT_remove_piston_from_mirror_map(map2,w, display_style);
# map3=FT_invert_mirror_map(map3, invert);

